/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, 1991, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /amnt/kona/tangram/u/william/X11/TGIF2/Chinese/RCS/cgif.c,v 1.37 1992/12/14 04:41:15 william Exp $";
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include "../const.h"
#include "../types.h"

#include "../cmd.e"
#include "../color.e"
#include "../cursor.e"
#include "../drawing.e"
#include "../dup.e"
#include "../file.e"
#include "../font.e"
#include "../grid.e"
#include "../mainloop.e"
#include "../mainmenu.e"
#include "../move.e"
#include "../msg.e"
#include "../names.e"
#include "../obj.e"
#include "../pattern.e"
#include "../raster.e"
#include "../select.e"
#include "../setup.e"
#include "../text.e"
#include "../xbitmap.e"

#include "spelling.xbm"

#define POPUP_NAME "ChineseKeyPad"

#define DEFAULT_CHAR_SIZE 16
#define CHINESE_ROWS 10
#define CHINESE_COLS 20
#define MAXCHARS 3

#define BUTTON_OK 0
#define BUTTON_FETCH 1
#define BUTTON_CANCEL 2
#define MAXBUTTONS 3

#define SPELL_CHINESE 0
#define SPELL_ENGLISH 1

#define CHINESE_RIGHT 0
#define CHINESE_DOWN 1

extern char	* getenv ();

int	lastFile = TRUE;

short	* pDrawFontAsc;
short	* pDrawFontDes;

/* chinese related variables */

static char	* firstTransTbl[] =
{  /* to be used if first character */
   "er",  "",    "",    "",
   "an",  "en",  "ang", "eng",
   "ai",  "ei",  "ao",  "ou",
   "a",   "o",   "e",   "e",
   "y",   "w",   "yu",  "",
   "z",   "c",   "s",   "",
   "zh",  "ch",  "sh",  "r",
   "j",   "q",   "x",   "",
   "g",   "k",   "h",   "",
   "d",   "t",   "n",   "l",
   "b",   "p",   "m",   "f",
};
static char	* transTbl[] =
{  /* to be used if not first character */
   "er",  "",    "",    "",
   "an",  "en",  "ang", "eng",
   "ai",  "ei",  "ao",  "ou",
   "a",   "o",   "e",   "e",
   "i",   "u",   "u",   "",
   "z",   "c",   "s",   "",
   "zh",  "ch",  "sh",  "r",
   "j",   "q",   "x",   "",
   "g",   "k",   "h",   "",
   "d",   "t",   "n",   "l",
   "b",   "p",   "m",   "f",
};

static int	charSize = DEFAULT_CHAR_SIZE;

static Window	baseWin;
static Window	dspWin;
static Window	scrollWin;

static struct BBRec	buttonBBox[MAXBUTTONS];
static struct BBRec	spellBBox;

static Pixmap	spellBitmap = None;
static Pixmap	zhuyinfuhaoBitmap = None;

static char	* buttonStr[MAXBUTTONS] = { "OK","FETCH","CANCEL" };

static int	dspWinX, dspWinY, dspWinH, dspWinW;
static int	baseWinW, baseWinH;
static int	scrollH = 0; /* dspWinH */

static int	dspY = 0; /* the top position of the current scroll */
static int	dspTotalH = 0; /* total height occupied by the objects */

static int	buttonStart = 0; /* x position of the title string */

static char	baseStr[MAXSTRING];
static int	baseStrStart = 0; /* x position of the title string */

static GC	chineseGC = NULL;
static GC	revChineseGC = NULL;
static GC	highlightGC = NULL;
static GC	zhuyinfuhaoGC = NULL;

static struct ObjRec	* topEntry = NULL, * botEntry = NULL;

static int	spellIndex = INVALID;
static int	spellChar[MAXCHARS];
static int	engSpellIndex = INVALID;
static char	engSpellChar[MAXSTRING];
static char	curName[MAXSTRING];

static char	chinesePath[MAXPATHLENGTH+1];
static int	chineseInit = FALSE;

static int	howSpelled = SPELL_CHINESE;

static Time		lastClickTime = 0;
static int		justClicked = FALSE;
static struct ObjRec	* lastSelected = NULL;

static
void DelAllEntries ()
{
   register struct ObjRec	* ptr;

   for (ptr = topEntry; ptr != NULL; ptr = ptr->next) FreeObj (ptr);
   topEntry = botEntry = NULL;
   dspTotalH = dspY = 0;
}

void CleanUpChinese ()
{
   if (chineseInit) XDestroyWindow (mainDisplay, baseWin);
   if (topEntry != NULL) DelAllEntries ();
   if (chineseGC != NULL) XFreeGC (mainDisplay, chineseGC);
   if (revChineseGC != NULL) XFreeGC (mainDisplay, revChineseGC);
   if (highlightGC != NULL) XFreeGC (mainDisplay, highlightGC);
   if (zhuyinfuhaoGC != NULL) XFreeGC (mainDisplay, zhuyinfuhaoGC);
   if (zhuyinfuhaoBitmap != NULL) XFreePixmap (mainDisplay, zhuyinfuhaoBitmap);
   chineseGC = NULL;
   revChineseGC = NULL;
   highlightGC = NULL;
   zhuyinfuhaoGC = NULL;
   zhuyinfuhaoBitmap = None;
   chineseInit = FALSE;
}

static
void DisplayXBm (ObjPtr, Normal)
   struct ObjRec	* ObjPtr;
   int			Normal;
{
   int		ltx, lty, w, h;
   XGCValues	values;

   ltx = ObjPtr->obbox.ltx; lty = ObjPtr->obbox.lty-dspY;
   w = ObjPtr->obbox.rbx-ltx; h = ObjPtr->obbox.rby-ObjPtr->obbox.lty;

   if (Normal)
   {
      XSetFillStyle (mainDisplay, revChineseGC, FillSolid);
      XFillRectangle (mainDisplay, dspWin, revChineseGC, ltx, lty, w, h);
      XSetFillStyle (mainDisplay, revChineseGC, FillStippled);

      values.ts_x_origin = ltx;
      values.ts_y_origin = lty;
      values.stipple = ObjPtr->detail.xbm->bitmap;
      XChangeGC (mainDisplay, chineseGC,
            GCTileStipXOrigin | GCTileStipYOrigin | GCStipple, &values);

      XFillRectangle (mainDisplay, dspWin, chineseGC, ltx, lty, w, h);
      XSetTSOrigin (mainDisplay, chineseGC, 0, 0);
   }
   else
   {
      XSetFillStyle (mainDisplay, chineseGC, FillSolid);
      XFillRectangle (mainDisplay, dspWin, chineseGC, ltx, lty, w, h);
      XSetFillStyle (mainDisplay, chineseGC, FillOpaqueStippled);

      values.ts_x_origin = ltx;
      values.ts_y_origin = lty;
      values.stipple = ObjPtr->detail.xbm->bitmap;
      XChangeGC (mainDisplay, revChineseGC,
            GCTileStipXOrigin | GCTileStipYOrigin | GCStipple, &values);

      XFillRectangle (mainDisplay, dspWin, revChineseGC, ltx, lty, w, h);
      XSetTSOrigin (mainDisplay, revChineseGC, 0, 0);
   }
   XSync (mainDisplay, False);
}

static
void LoadDisplay (FileName)
   char	* FileName;
{
   FILE			* fp;
   int			len, read_status, cur_x, cur_y, cur_h, dx, dy;
   char 		full_name[MAXPATHLENGTH], msg[MAXSTRING];
   struct ObjRec	* obj_ptr, * saved_top_obj, * saved_bot_obj;

   len = strlen (FileName);
   if (len >= 4 && strcmp (&FileName[len-4], ".obj") == 0)
   {
      if (*FileName == '/' || *FileName == '.')
         strcpy (full_name, FileName);
      else
         sprintf (full_name, "%s/%s", chinesePath, FileName);
   }
   else
   {
      if (*FileName == '/' || *FileName == '.')
         sprintf (full_name, "%s.obj", FileName);
      else
         sprintf (full_name, "%s/%s.obj", chinesePath, FileName);
   }

   if ((fp = fopen (full_name, "r")) == NULL)
   {
      sprintf (msg, "Can not load file '%s'.", full_name);
      Msg (msg);
      return;
   }

   saved_top_obj = topObj;
   saved_bot_obj = botObj;
   topObj = botObj = NULL;

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);
   SetWatchCursor (baseWin);

   importingFile = TRUE;
   while ((read_status = ReadObj (fp, &obj_ptr, FALSE)) == TRUE)
      if (obj_ptr != NULL)
      {
         AdjForOldVersion (obj_ptr);
         if (obj_ptr->type == OBJ_XBM)
            AddObj (NULL, topObj, obj_ptr);
         else
            FreeObj (obj_ptr);
      }
   fclose (fp);
   importingFile = FALSE;

   if (read_status == INVALID)
   {
      sprintf (msg, "File version too large (=%1d).  Load aborted!",
            fileVersion);
      Msg (msg);
      SetDefaultCursor (mainWindow);
      SetDefaultCursor (drawWindow);
      return;
   }

   SetDefaultCursor (baseWin);
   SetDefaultCursor (mainWindow);
   SetDefaultCursor (drawWindow);

   cur_x = 0; cur_y = 0; cur_h = 0;
   for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      int	w, h, win_w = dspWinW-2*brdrW;

      w = obj_ptr->obbox.rbx - obj_ptr->obbox.ltx;
      h = obj_ptr->obbox.rby - obj_ptr->obbox.lty;
      if (cur_x+w > win_w)
      {
         cur_x = 0;
         cur_y += cur_h;
         cur_h = 0;
      }
      if (cur_x+w > win_w)
      {
         sprintf (msg, "Object width (%1d) > dspWinW (%1d)!  Fetch Aborted!",
               w, win_w);
         Msg (msg);
         DelAllObj ();
         topObj = saved_top_obj;
         botObj = saved_bot_obj;
      }
      if (h > cur_h) cur_h = h;
      dx = cur_x - obj_ptr->obbox.ltx;
      dy = cur_y - obj_ptr->obbox.lty;
      MoveObj (obj_ptr, dx, dy);
      cur_x += w;
   }

   DelAllEntries ();

   dspY = 0;
   dspTotalH = cur_y+cur_h;
/*    dspTotalH = (((int)(dspTotalH/charSize))+1)*charSize; */
/*    dspTotalH = (((int)(dspTotalH/charSize))+1)*charSize; */

   topEntry = topObj;
   botEntry = botObj;
   topObj = saved_top_obj;
   botObj = saved_bot_obj;
}

static
void RedrawScrollWin ()
{
   double	frac, start_frac;
   int		block_h, block_start;
   XGCValues	values;

   XClearWindow (mainDisplay, scrollWin);

   start_frac = (dspTotalH > 0) ?
         (double)((double)(dspY) / (double)(dspTotalH)) : ((double)0.0);
   block_start = (int)(scrollH * start_frac); /* starting pixel */

   if (dspTotalH > scrollH)
      frac = (double)((double)scrollH / (double)(dspTotalH));
   else
      frac = 1.0;

   if (dspY+scrollH >= dspTotalH)
      block_h = scrollH - block_start;
   else
      block_h = (int)(scrollH * frac);

   values.foreground = myBgPixel;
   values.background = myFgPixel;
   values.function = GXcopy;
   values.fill_style = FillSolid;
   XChangeGC (mainDisplay, patGC,
         GCForeground | GCBackground | GCFunction | GCFillStyle, &values);
   XFillRectangle (mainDisplay, scrollWin, patGC, 0, 0,
         scrollBarW, scrollH);

   values.foreground = myFgPixel;
   values.background = myBgPixel;
   values.fill_style = FillOpaqueStippled;
   values.stipple = patPixmap[SCROLLPAT];
   XChangeGC (mainDisplay, patGC,
         GCForeground | GCBackground | GCFillStyle | GCStipple, &values);
   XFillRectangle (mainDisplay, scrollWin, patGC, 0, block_start,
         scrollBarW, block_h);
}

static
void RedrawDspWindow ()
{
   register struct ObjRec	* obj_ptr;
   int				ltx, lty, w, h;
   XGCValues			values;

   XClearWindow (mainDisplay, dspWin);

   for (obj_ptr = botEntry; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      ltx = obj_ptr->obbox.ltx;
      lty = obj_ptr->obbox.lty;
      w = obj_ptr->obbox.rbx - ltx;
      h = obj_ptr->obbox.rby - lty;

      if (lty > dspY+scrollH) break;
      if (lty+h < dspY) continue;

      values.ts_x_origin = ltx;
      values.ts_y_origin = lty-dspY;
      values.stipple = obj_ptr->detail.xbm->bitmap;
      XChangeGC (mainDisplay, chineseGC,
            GCTileStipXOrigin | GCTileStipYOrigin | GCStipple, &values);

      XFillRectangle (mainDisplay, dspWin, chineseGC, ltx, lty-dspY, w, h);
   }
   XSetTSOrigin (mainDisplay, chineseGC, 0, 0);
   if (lastSelected != NULL) DisplayXBm (lastSelected, BUTTON_INVERT);
}

static
void ReverseZhuYenFuHao (index, x, y)
   int	index, * x, * y;
{
   *y = (index%4)*DEFAULT_CHAR_SIZE;
   *x = ((int)(index/4))*DEFAULT_CHAR_SIZE;
}

static
void RedrawName ()
{
   register int	i, y;
   int		len;
   char		* c_ptr;
   XGCValues	values;

   XClearArea (mainDisplay, baseWin, dspWinX, dspWinY-2*defaultFontHeight,
         dspWinW+4*brdrW, defaultFontHeight+4, False);
   XDrawRectangle (mainDisplay, baseWin, defaultGC, dspWinX,
         dspWinY-2*defaultFontHeight, dspWinW+4*brdrW+4, defaultFontHeight+4);

   XClearArea (mainDisplay, baseWin, dspWinX-1,
         dspWinY-3*defaultFontHeight-3*DEFAULT_CHAR_SIZE-1,
         DEFAULT_CHAR_SIZE+2, 3*DEFAULT_CHAR_SIZE+2, False);
   XDrawRectangle (mainDisplay, baseWin, defaultGC, dspWinX-1,
         dspWinY-3*defaultFontHeight-3*DEFAULT_CHAR_SIZE-1,
         DEFAULT_CHAR_SIZE+2, 3*DEFAULT_CHAR_SIZE+2);

   switch (howSpelled)
   {
      case SPELL_CHINESE:
         if (spellIndex == 0) return;
         strcpy (curName, firstTransTbl[spellChar[0]]);

         for (i = 1; i < spellIndex; i++)
         {
            strcat (curName, "-");
            strcat (curName, transTbl[spellChar[i]]);
         }

         y = dspWinY-3*defaultFontHeight-3*DEFAULT_CHAR_SIZE;
         for (i = 0; i < spellIndex; i++, y+=DEFAULT_CHAR_SIZE)
         {
            int	src_x, src_y;

            ReverseZhuYenFuHao (spellChar[i], &src_x, &src_y);
            XCopyArea (mainDisplay, spellBitmap, zhuyinfuhaoBitmap,
                  zhuyinfuhaoGC, src_x, src_y,
                  DEFAULT_CHAR_SIZE, DEFAULT_CHAR_SIZE, 0, 0);

            values.fill_style = FillStippled;
            values.ts_x_origin = dspWinX;
            values.ts_y_origin = y;
            values.stipple = zhuyinfuhaoBitmap;;
            XChangeGC (mainDisplay, chineseGC,
                  GCFillStyle | GCTileStipXOrigin |
                  GCTileStipYOrigin | GCStipple, &values);

            XFillRectangle (mainDisplay, baseWin, chineseGC, dspWinX, y,
                  DEFAULT_CHAR_SIZE, DEFAULT_CHAR_SIZE);
         }
         XSetFillStyle (mainDisplay, chineseGC, FillSolid);
         XSetTSOrigin (mainDisplay, chineseGC, 0, 0);
         XDrawString (mainDisplay, baseWin, chineseGC, dspWinX+2,
               dspWinY-defaultFontHeight-defaultFontDes+1, curName,
               strlen(curName));
         XSetFillStyle (mainDisplay, chineseGC, FillOpaqueStippled);
         break;
      case SPELL_ENGLISH:
         if (engSpellIndex == 0) return;
         strcpy (curName, engSpellChar);
         len = strlen (curName);
         if (len*defaultFontWidth+5 > dspWinW+4*brdrW+4)
         {
            int	max_len = (int)((dspWinW+4*brdrW-1)/defaultFontWidth);

            c_ptr = &(curName[len-max_len]);
            len = max_len;
         }
         else
            c_ptr = curName;
         XSetFillStyle (mainDisplay, chineseGC, FillSolid);
         XDrawString (mainDisplay, baseWin, chineseGC, dspWinX+2,
               dspWinY-defaultFontHeight-defaultFontDes+1, c_ptr,
               len);
         XDrawLine (mainDisplay, baseWin, chineseGC,
               dspWinX+2+len*defaultFontWidth, dspWinY-2*defaultFontHeight+1,
               dspWinX+2+len*defaultFontWidth, dspWinY-defaultFontHeight+1);
         XSetFillStyle (mainDisplay, chineseGC, FillOpaqueStippled);
         break;
   }
}

static
void RedrawButtons ()
{
   register int	i, x;

   x = buttonStart;
   for (i = 0; i < MAXBUTTONS; i++)
   {
      buttonBBox[i].lty = dspWinY+dspWinH+defaultFontHeight;
      buttonBBox[i].ltx = x;
      DisplayButton (baseWin, buttonStr[i], 8, &(buttonBBox[i]),
            BUTTON_NORMAL);
      x = buttonBBox[i].rbx + 1 + defaultFontWidth;
   }
}

static
void RedrawBaseWin ()
{
   XGCValues	values;

   XClearWindow (mainDisplay, baseWin);

   XDrawRectangle (mainDisplay, baseWin, defaultGC, 0, 0, baseWinW-1,
         baseWinH-1);
   XDrawString (mainDisplay, baseWin, defaultGC, baseStrStart,
         defaultFontHeight+defaultFontAsc+2, baseStr, strlen(baseStr));

   values.ts_x_origin = spellBBox.ltx;
   values.ts_y_origin = spellBBox.lty;
   values.stipple = spellBitmap;
   XChangeGC (mainDisplay, chineseGC,
         GCTileStipXOrigin | GCTileStipYOrigin | GCStipple, &values);

   XFillRectangle (mainDisplay, baseWin, chineseGC, spellBBox.ltx,
         spellBBox.lty, spellBBox.rbx-spellBBox.ltx,
         spellBBox.rby-spellBBox.lty);
   XSetTSOrigin (mainDisplay, chineseGC, 0, 0);
   XDrawRectangle (mainDisplay, baseWin, defaultGC, spellBBox.ltx-1,
         spellBBox.lty-1, spellBBox.rbx-spellBBox.ltx+2,
         spellBBox.rby-spellBBox.lty+2);

   RedrawButtons ();
   RedrawName ();
}

static
void ScrollHandler (button_ev)
   XButtonEvent	* button_ev;
{
   double	frac, start_frac;
   int		block_h, block_start;

   if (button_ev->button == Button3)
   {
      if (dspY == 0) return;

      dspY -= charSize;
      if (dspY < 0) dspY = 0;
   }
   else if (button_ev->button == Button1)
   {
      if (dspY + scrollH >= dspTotalH) return;

      dspY += charSize;
   }
   else if (button_ev->button == Button2)
   {  /* recall that scrollH and dspTotalH are all multiples of charSize */
      int	done=FALSE;
      XEvent	ev;

      if (dspTotalH <= scrollH) return;

      frac = (double)((double)scrollH / (double)dspTotalH);
      block_h = (int)(scrollH * frac);
      block_start = button_ev->y;
      start_frac = (double)((double)block_start / (double)scrollH);
      if (block_start+block_h >= scrollH)
         dspY = dspTotalH - scrollH;
      else
         dspY = (int)(dspTotalH * start_frac);

      RedrawScrollWin ();
      RedrawDspWindow ();
      XGrabPointer (mainDisplay, scrollWin, False,
            PointerMotionMask | ButtonReleaseMask, GrabModeAsync,
            GrabModeAsync, None, handCursor, CurrentTime);

      while (!done)
      {
         XNextEvent (mainDisplay, &ev);

         if (ev.type == Expose || ev.type == VisibilityNotify)
            ExposeEventHandler (&ev, TRUE);
         else if (ev.type == ButtonRelease)
         {
            XUngrabPointer (mainDisplay, CurrentTime);
            done = TRUE;
         }
         else if (ev.type == MotionNotify)
         {
            int new_name_first, y=ev.xmotion.y;

            start_frac = (double)(((double)y) / ((double)scrollH));

            if (y <= 0)
               new_name_first = 0;
            else if (y+block_h >= scrollH)
               new_name_first = dspTotalH - scrollH;
            else
               new_name_first = (int)(dspTotalH * start_frac);

            if (dspY != new_name_first)
            {
               dspY = new_name_first;
               RedrawScrollWin ();
               RedrawDspWindow ();
            }
            while (XCheckMaskEvent (mainDisplay, PointerMotionMask, &ev)) ;
         }
      }
      return;
   }
   RedrawScrollWin ();
   RedrawDspWindow ();
}

static
struct ObjRec * DspHandler (button_ev)
   XButtonEvent	* button_ev;
{
   register struct BBRec	* bbox_ptr;
   register int			x, y;
   Time				click_time;
   struct ObjRec		* obj_ptr;

   x = button_ev->x;
   y = button_ev->y+dspY;

   for (obj_ptr = botEntry; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      bbox_ptr = &(obj_ptr->obbox);
      if (x >= bbox_ptr->ltx && x <= bbox_ptr->rbx &&
            y >= bbox_ptr->lty && y <= bbox_ptr->rby)
         break;
   }

   if (obj_ptr != NULL)
   {
      if (lastSelected != NULL) DisplayXBm (lastSelected, BUTTON_NORMAL);
      DisplayXBm (obj_ptr, BUTTON_INVERT);
   }
   else if (lastSelected)
      DisplayXBm (lastSelected, BUTTON_NORMAL);

   click_time = button_ev->time;
   if (justClicked && obj_ptr!=NULL && lastSelected==obj_ptr &&
         (click_time-lastClickTime)<doubleClickInterval)
      return (obj_ptr);

   justClicked = TRUE;
   lastClickTime = click_time;
   lastSelected = obj_ptr;
   return (NULL);
}

static
int GetZhuYenFuHao (button_ev, ltx, lty)
   XButtonEvent	* button_ev;
   int		* ltx, * lty;
{
   int	x = button_ev->x-spellBBox.ltx, y = button_ev->y-spellBBox.lty;
   int	row, col;

   row = (int)(y/DEFAULT_CHAR_SIZE);
   col = (int)(x/DEFAULT_CHAR_SIZE);
   *ltx = col*DEFAULT_CHAR_SIZE + spellBBox.ltx;
   *lty = row*DEFAULT_CHAR_SIZE + spellBBox.lty;
   return (col*4+row);
}

static
void HandleControlChar (key_sym)
   KeySym	key_sym;
{
   struct ObjRec	* old_selected = lastSelected;

   switch (key_sym)
   {
      case XK_Left: key_sym = ((unsigned long)'p')&0xff; break;
      case XK_Up: return;
      case XK_Right: key_sym = ((unsigned long)'n')&0xff; break;
      case XK_Down: return;
   }
   switch (key_sym&0xff)
   {
      case 'n': /* next one */
      case 'f':
         if (lastSelected == NULL || lastSelected->prev == NULL)
            lastSelected = botEntry;
         else
            lastSelected = lastSelected->prev;
         break;
      case 'p': /* previous one */
      case 'b':
         if (lastSelected == NULL || lastSelected->next == NULL)
            lastSelected = topEntry;
         else
            lastSelected = lastSelected->next;
         break;
      default: return;
   }
   if (old_selected != NULL) DisplayXBm (old_selected, BUTTON_NORMAL);
   DisplayXBm (lastSelected, BUTTON_INVERT);
   justClicked = FALSE;
}

static
void HandleBackSpace (just_fetched, exposed)
   int	* just_fetched, exposed;
{
   int	need_redraw;

   switch (howSpelled)
   {
      case SPELL_CHINESE:
         if (spellIndex != 0)
         {
            need_redraw = *just_fetched;

            *just_fetched = FALSE;
            --spellIndex;

            if (topEntry != NULL) DelAllEntries ();
            lastSelected = NULL;
            if (exposed)
            {
               RedrawName ();
               if (need_redraw)
               {
                  RedrawScrollWin ();
                  RedrawDspWindow ();
               }
            }
         }
         break;
      case SPELL_ENGLISH:
         if (engSpellIndex != 0)
         {
            need_redraw = *just_fetched;

            *just_fetched = FALSE;
            --engSpellIndex;
            engSpellChar[engSpellIndex] = '\0';

            if (topEntry != NULL) DelAllEntries ();
            lastSelected = NULL;
            if (exposed)
            {
               RedrawName ();
               if (need_redraw)
               {
                  RedrawScrollWin ();
                  RedrawDspWindow ();
               }
            }
         }
         break;
   }
}

static
void FetchClicked (just_fetched, exposed)
   int	* just_fetched, exposed;
{
   DisplayButton (baseWin, buttonStr[BUTTON_FETCH], 8,
         &buttonBBox[BUTTON_FETCH], BUTTON_INVERT);
   XSync (mainDisplay, False);
   DisplayButton (baseWin, buttonStr[BUTTON_FETCH], 8,
         &buttonBBox[BUTTON_FETCH], BUTTON_NORMAL);
   XSync (mainDisplay, False);
   *just_fetched = TRUE;
   if (howSpelled==SPELL_CHINESE && spellIndex!=0 ||
         howSpelled==SPELL_ENGLISH && engSpellIndex!=0)
      LoadDisplay (curName);
   lastSelected = NULL;
   if (exposed)
   {
      RedrawName ();
      RedrawScrollWin ();
      RedrawDspWindow ();
   }
}

static int	chineseWinExposed = FALSE;
static int	justFetched = FALSE;
static int	extraWinIndex = INVALID;
static int	chineseDirection = CHINESE_RIGHT;
static struct ObjRec	* savedSelected = NULL;

static
void ChineseWinExposeHandler (input)
   XEvent	* input;
{
   XEvent	ev;

   while (XCheckWindowEvent (mainDisplay,baseWin,ExposureMask,&ev)) ;
   while (XCheckWindowEvent (mainDisplay,scrollWin,ExposureMask, &ev)) ;
   while (XCheckWindowEvent (mainDisplay,dspWin,ExposureMask,&ev)) ;
   while (XCheckWindowEvent (mainDisplay,baseWin,StructureNotifyMask, &ev)) ;

   RedrawBaseWin ();
   RedrawScrollWin ();
   RedrawDspWindow ();

   chineseWinExposed = TRUE;
   XSync (mainDisplay, False);
}

static
void DoneSelectChineseChar (obj_ptr, button_selected)
   struct ObjRec	* obj_ptr;
   int			button_selected;
{
   struct ObjRec	* new_obj_ptr;
   struct XBmRec	* xbm_ptr, * new_xbm_ptr;
   Pixmap		bitmap;
   int			keep_typing, w, h;
   char			msg[MAXSTRING];

   if (chineseWinExposed)
   {
      DisplayButton (baseWin, buttonStr[button_selected], 8,
            &buttonBBox[button_selected], BUTTON_INVERT);
      XSync (mainDisplay, False);
      DisplayButton (baseWin, buttonStr[button_selected], 8,
            &buttonBBox[button_selected], BUTTON_NORMAL);
      XSync (mainDisplay, False);
   }

   keep_typing = (topSel!=NULL && topSel==botSel && topSel->obj==savedSelected);

   TieLooseEnds ();
   SetCurChoice (NOTHING);
   if (topSel != NULL) { HighLightReverse (); RemoveAllSel (); }

   if (obj_ptr == NULL)
   {
      if (button_selected == BUTTON_CANCEL)
      {
         XUnmapWindow (mainDisplay, baseWin);
         extraWinInfo[extraWinIndex].mapped = FALSE;
         chineseWinExposed = FALSE;
      }
      return;
   }

   xbm_ptr = obj_ptr->detail.xbm;

   w = obj_ptr->obbox.rbx - obj_ptr->obbox.ltx;
   h = obj_ptr->obbox.rby - obj_ptr->obbox.lty;

   bitmap = XCreatePixmap (mainDisplay, mainWindow, w, h, 1);
   if (bitmap == None)
   {
      sprintf (msg, "Can not create bitmap of size %1dx%1d.", w, h);
      Msg (msg);
      return;
   }
   XCopyArea (mainDisplay, xbm_ptr->bitmap, bitmap, xbmGC, 0, 0, w, h, 0, 0);

   new_obj_ptr = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));
   DupObjBasics (obj_ptr, new_obj_ptr);
   new_obj_ptr->color = colorIndex;

   new_xbm_ptr = (struct XBmRec *) calloc (1, sizeof(struct XBmRec));
   new_obj_ptr->detail.xbm = new_xbm_ptr;

   new_xbm_ptr->image = NULL;
   new_xbm_ptr->data = NULL;
   new_xbm_ptr->fill = objFill;
   new_xbm_ptr->image_w = xbm_ptr->image_w;
   new_xbm_ptr->image_h = xbm_ptr->image_h;
   new_xbm_ptr->rotate = ROTATE0;
   new_xbm_ptr->flip = 0;
   new_xbm_ptr->cached_zoom = 0;
   new_xbm_ptr->cached_rotate = INVALID;
   new_xbm_ptr->cached_flip = 0;
   new_xbm_ptr->cached_w = new_xbm_ptr->cached_h = 0;

   new_xbm_ptr->bitmap = bitmap;
   new_xbm_ptr->cached_bitmap = None;

   new_xbm_ptr->real_type = xbm_ptr->real_type;

   new_obj_ptr->obbox.ltx = obj_ptr->obbox.ltx;
   new_obj_ptr->obbox.lty = obj_ptr->obbox.lty;
   new_obj_ptr->obbox.rbx = new_obj_ptr->bbox.rbx = obj_ptr->obbox.ltx+w;
   new_obj_ptr->obbox.rby = new_obj_ptr->bbox.rby = obj_ptr->obbox.lty+h;

   AdjObjBBox (new_obj_ptr);
   if (keep_typing)
   {
      int	x, y, dx, dy;

      x = savedSelected->obbox.ltx;
      y = savedSelected->obbox.lty;
      switch (chineseDirection)
      {
         case CHINESE_RIGHT:
            x = savedSelected->obbox.rbx;
            y = savedSelected->obbox.lty;
            break;
         case CHINESE_DOWN:
            x = savedSelected->obbox.ltx;
            y = savedSelected->obbox.rby+textVSpace;
            break;
      }

      dx = x - new_obj_ptr->obbox.ltx;
      dy = y - new_obj_ptr->obbox.lty;
      MoveObj (new_obj_ptr, dx, dy);
      DrawObj (drawWindow, new_obj_ptr);
   }
   else
   {
      XWarpPointer (mainDisplay, None, drawWindow, 0, 0, 0, 0,
            (int)(drawWinW>>(zoomScale+1)), (int)(drawWinH>>(zoomScale+1)));
      PlaceTopObj (new_obj_ptr);
   }

   AdjObjSplineVs (new_obj_ptr);
   AddObj (NULL, topObj, new_obj_ptr);
   RecordNewObjCmd ();
   SelectTopObj ();
   SetFileModified (TRUE);
   justDupped = FALSE;

   savedSelected = topObj;
}

static
int ChineseWinEvHandler (input)
   XEvent	* input;
{
   int			index = 0;
   XEvent		ev;
   char			buf[80];
   XKeyEvent		* key_ev;
   XButtonEvent		* button_ev;
   KeySym		key_sym;

   if ((input->type==MapNotify && input->xany.window==baseWin) ||
         (input->type==Expose && (input->xany.window==baseWin ||
         input->xany.window==scrollWin ||
         input->xany.window==dspWin)) ||
         (!chineseWinExposed &&
         (XCheckWindowEvent (mainDisplay,baseWin,ExposureMask,&ev) ||
         XCheckWindowEvent (mainDisplay,scrollWin,ExposureMask,&ev) ||
         XCheckWindowEvent (mainDisplay,dspWin,ExposureMask,&ev) ||
         XCheckWindowEvent (mainDisplay,baseWin,StructureNotifyMask,
         &ev))))
   {
      while (XCheckWindowEvent (mainDisplay,baseWin,ExposureMask,&ev)) ;
      while (XCheckWindowEvent (mainDisplay,scrollWin,ExposureMask, &ev)) ;
      while (XCheckWindowEvent (mainDisplay,dspWin,ExposureMask,&ev)) ;
      while (XCheckWindowEvent (mainDisplay,baseWin,StructureNotifyMask, &ev)) ;

      RedrawBaseWin ();
      RedrawScrollWin ();
      RedrawDspWindow ();

      chineseWinExposed = TRUE;
      XSync (mainDisplay, False);

      if ((input->type==MapNotify && input->xany.window==baseWin) ||
            (input->type==Expose && (input->xany.window==baseWin ||
            input->xany.window==scrollWin ||
            input->xany.window==dspWin)))
         return (INVALID);
   }

   if (input->type==Expose)
      ChineseWinExposeHandler (input);
   else if (input->type == KeyPress)
   {
      key_ev = &(input->xkey);
      XLookupString (key_ev, buf, 80, &key_sym, NULL);

      if ((buf[0]=='\r' && (key_sym & 0xff)=='\r') ||
          (buf[0]=='\n' && (key_sym & 0xff)=='\n'))
      {
         if (justFetched)
            DoneSelectChineseChar (lastSelected, BUTTON_OK);
         else
         {
            DisplayButton (baseWin, buttonStr[BUTTON_FETCH], 8,
                  &buttonBBox[BUTTON_FETCH], BUTTON_INVERT);
            XSync (mainDisplay, False);
            DisplayButton (baseWin, buttonStr[BUTTON_FETCH], 8,
                  &buttonBBox[BUTTON_FETCH], BUTTON_NORMAL);
            XSync (mainDisplay, False);
            justFetched = TRUE;
            if (howSpelled==SPELL_CHINESE && spellIndex!=0 ||
                  howSpelled==SPELL_ENGLISH && engSpellIndex!=0)
               LoadDisplay (curName);
            lastSelected = NULL;
            if (chineseWinExposed)
            {
               RedrawName ();
               RedrawScrollWin ();
               RedrawDspWindow ();
            }
         }
      }
      else if (buf[0]=='\033' && (key_sym & 0xff)=='\033')
         DoneSelectChineseChar (NULL, BUTTON_CANCEL);
      else if ((buf[0] == '\b'  && (key_sym & 0xff)=='\b') ||
            (buf[0] == '\b' && (key_sym & 0xff)=='h') ||
            (buf[0] == '\177' && (key_sym & 0x7f)=='\177'))
         HandleBackSpace (&justFetched, chineseWinExposed);
      else if (chineseWinExposed && topEntry != NULL &&
            ((key_sym>'\040' && key_sym<='\177' &&
            (key_ev->state & ControlMask)) || key_sym==XK_Up ||
            key_sym==XK_Down || key_sym==XK_Left || key_sym==XK_Right))
         HandleControlChar (key_sym);
      else if (key_sym>='\040' && key_sym<='\177')
      {
         if (howSpelled == SPELL_CHINESE)
         {
            howSpelled = SPELL_ENGLISH;
            engSpellIndex = 0;
            engSpellChar[0] = '\0';
         }
         if (justFetched)
         {
            engSpellIndex = 0;
            if (topEntry != NULL) DelAllEntries ();
            justFetched = FALSE;
            lastSelected = NULL;
            if (chineseWinExposed)
            {
               RedrawScrollWin ();
               RedrawDspWindow ();
            }
         }
         engSpellChar[engSpellIndex++] = buf[0];
         engSpellChar[engSpellIndex] = '\0';
         if (chineseWinExposed) RedrawName ();
      }
   }
   else if (input->type == ButtonPress)
   {
      button_ev = &(input->xbutton);
      if (button_ev->window == baseWin)
      {
         if (button_ev->button == Button2)
            FetchClicked (&justFetched, chineseWinExposed);
         else if (button_ev->button == Button3)
            HandleBackSpace (&justFetched, chineseWinExposed);
         else if (PointInBBox (button_ev->x, button_ev->y,
               buttonBBox[BUTTON_OK]))
            DoneSelectChineseChar (lastSelected, BUTTON_OK);
         else if (PointInBBox (button_ev->x, button_ev->y,
               buttonBBox[BUTTON_FETCH]))
            FetchClicked (&justFetched, chineseWinExposed);
         else if (PointInBBox (button_ev->x, button_ev->y,
               buttonBBox[BUTTON_CANCEL]))
            DoneSelectChineseChar (NULL, BUTTON_CANCEL);
         else if (PointInBBox (button_ev->x, button_ev->y, spellBBox))
         {  /* click in the spelling area */
            if (howSpelled == SPELL_ENGLISH)
            {
               howSpelled = SPELL_CHINESE;
               spellIndex = 0;
            }
            if (justFetched)
            {
               spellIndex = 0;
               if (topEntry != NULL) DelAllEntries ();
               justFetched = FALSE;
               lastSelected = NULL;
               if (chineseWinExposed)
               {
                  RedrawScrollWin ();
                  RedrawDspWindow ();
               }
            }
            if (spellIndex != MAXCHARS)
            {
               int		ltx, lty;

               index = GetZhuYenFuHao (button_ev, &ltx, &lty);
               if (*firstTransTbl[index] != '\0')
               {
                  spellChar[spellIndex++] = index;

                  XFillRectangle (mainDisplay, baseWin, highlightGC,
                        ltx, lty, DEFAULT_CHAR_SIZE, DEFAULT_CHAR_SIZE);
                  XSync (mainDisplay, False);
                  XFillRectangle (mainDisplay, baseWin, highlightGC,
                        ltx, lty, DEFAULT_CHAR_SIZE, DEFAULT_CHAR_SIZE);
                  XSync (mainDisplay, False);
               }

               lastSelected = NULL;
               if (chineseWinExposed) RedrawName ();
            }
         }
      }
      else if (button_ev->window == scrollWin)
         ScrollHandler (button_ev);
      else if (button_ev->window == dspWin)
      {
         if (button_ev->button == Button2)
            FetchClicked (&justFetched, chineseWinExposed);
         else if (button_ev->button == Button3)
            HandleBackSpace (&justFetched, chineseWinExposed);
         else if (DspHandler (button_ev) != NULL)
            DoneSelectChineseChar (lastSelected, BUTTON_OK);
      }
   }
   return (INVALID);
}

static
void InitChinese ()
{
   int		default_found = FALSE;
   int		graph_start;
   int		str_width, button_widths, graph_width;
   int		dsp_w, dsp_h, w, h, win_x, win_y;
   char		* c_ptr;
   XWMHints	wmhints;
   XSizeHints	sizehints;
   XGCValues	values;
   XSetWindowAttributes	win_attrs;

   chineseDirection = CHINESE_RIGHT;
   if ((c_ptr = XGetDefault(mainDisplay,TOOL_NAME,"ChineseDirection")) != NULL)
      if (strcmp (c_ptr, "down") == 0 || strcmp (c_ptr, "Down") == 0)
         chineseDirection = CHINESE_DOWN;

   if ((c_ptr = XGetDefault(mainDisplay, TOOL_NAME,
         "ChineseCharHeight")) != NULL)
      charSize = atoi (c_ptr);

   if ((c_ptr = getenv ("TGIF_CHINESE_PATH")) == NULL)
      strcpy (chinesePath, TGIF_CPATH);
   else
      if (strlen (c_ptr) >= MAXPATHLENGTH)
         strcpy (chinesePath, TGIF_CPATH);
      else
         strcpy (chinesePath, c_ptr);

   values.foreground = myFgPixel;
   values.background = myBgPixel;
   values.fill_style = FillOpaqueStippled;
   values.font = defaultFontPtr->fid;
   chineseGC = XCreateGC (mainDisplay, rootWindow,
         GCForeground | GCBackground | GCFillStyle | GCFont, &values);

   values.foreground = myBgPixel;
   values.background = myFgPixel;
   values.fill_style = FillStippled;
   revChineseGC = XCreateGC (mainDisplay, rootWindow,
         GCForeground | GCBackground | GCFillStyle, &values);

   values.foreground = xorOne;
   values.function = GXxor;
   values.fill_style = FillSolid;
   highlightGC = XCreateGC (mainDisplay, rootWindow,
         GCForeground | GCFunction | GCFillStyle, &values);

   zhuyinfuhaoBitmap = XCreatePixmap (mainDisplay, rootWindow,
         DEFAULT_CHAR_SIZE, DEFAULT_CHAR_SIZE, 1);

   values.foreground = 0;
   values.background = 0;
   values.function = GXcopy;
   values.fill_style = FillSolid;
   zhuyinfuhaoGC = XCreateGC (mainDisplay, zhuyinfuhaoBitmap,
         GCForeground | GCBackground | GCFunction | GCFillStyle, &values);

   XFillRectangle (mainDisplay, zhuyinfuhaoBitmap, zhuyinfuhaoGC, 0, 0,
         DEFAULT_CHAR_SIZE, DEFAULT_CHAR_SIZE);
   XSetForeground (mainDisplay, zhuyinfuhaoGC, 1);

   dspWinW = CHINESE_COLS * charSize + 2*brdrW;
   dspWinH = CHINESE_ROWS * charSize + 2*brdrW;
   scrollH = CHINESE_ROWS * charSize;

   spellBitmap = XCreateBitmapFromData (mainDisplay, mainWindow,
        spelling_bits, spelling_width, spelling_height);

   dsp_w = DisplayWidth (mainDisplay, mainScreen);
   dsp_h = DisplayHeight (mainDisplay, mainScreen);

   button_widths = ButtonWidth("OK", 8) + ButtonWidth("FETCH", 8) +
         ButtonWidth("CANCEL", 8) + 2*(defaultFontWidth+1);

   strcpy (baseStr, "Please select using zhu-yin-fu-hao...");

   str_width = defaultFontWidth * strlen (baseStr);
   graph_width = dspWinW + scrollBarW + 2 * brdrW;

   if (str_width > graph_width)
   {
      w = str_width + 4 * defaultFontWidth;
      baseStrStart = 2 * defaultFontWidth;
      graph_start = (w - graph_width) / 2;
   }
   else
   {
      w = graph_width + 4 * defaultFontWidth;
      baseStrStart = (w - str_width) / 2;
      graph_start = 2 * defaultFontWidth;
   }

   spellBBox.ltx = (w - spelling_width) / 2;
   spellBBox.lty = 3*defaultFontHeight;
   spellBBox.rbx = spellBBox.ltx+spelling_width;
   spellBBox.rby = spellBBox.lty+spelling_height;

   buttonStart = (w - button_widths) / 2;
   h = 9*defaultFontHeight + spelling_height + dspWinH;

   win_x = (w > dsp_w) ? 0 : (dsp_w - w)/2;
   win_y = (h > dsp_h) ? 0 : (dsp_h - h)/3;

   baseWinW = w;
   baseWinH = h;
   if ((baseWin = XCreateSimpleWindow (mainDisplay, rootWindow,
         win_x, win_y, w, h, brdrW, myBorderPixel, myBgPixel)) == 0)
      Error ("InitChinese()", "Can not XCreateSimpleWindow() for baseWin");

   XDefineCursor (mainDisplay, baseWin, defaultCursor);

   dspWinX = graph_start;
   dspWinY = 6*defaultFontHeight+spelling_height;
   if ((dspWin = XCreateSimpleWindow (mainDisplay, baseWin, dspWinX, dspWinY,
         dspWinW-2*brdrW, dspWinH-2*brdrW, brdrW,
         myBorderPixel, myBgPixel)) == 0)
      Error ("InitChinese()", "Can not XCreateSimpleWindow() for dspWin");

   if ((scrollWin = XCreateSimpleWindow (mainDisplay, baseWin,
         graph_start+dspWinW, 6*defaultFontHeight+spelling_height,
         scrollBarW, dspWinH-2*brdrW, brdrW, myBorderPixel, myBgPixel)) == 0)
      Error ("InitChinese()","Can not XCreateSimpleWindow() for scrollWin");

   win_attrs.save_under = True;
   XChangeWindowAttributes (mainDisplay, baseWin, CWSaveUnder, &win_attrs);

   wmhints.flags = InputHint | StateHint;
   wmhints.input = True;
   wmhints.initial_state = NormalState;
   XSetWMHints (mainDisplay, baseWin, &wmhints);

   sizehints.flags = PPosition | PSize | USPosition | PMinSize | PMaxSize;
   sizehints.x = win_x;
   sizehints.y = win_y;
   sizehints.width = sizehints.min_width = sizehints.max_width = w;
   sizehints.height = sizehints.min_height = sizehints.max_height = h;
#ifdef NOTR4MODE
   XSetNormalHints (mainDisplay, baseWin, &sizehints);
#else
   XSetWMNormalHints (mainDisplay, baseWin, &sizehints);
#endif
   XStoreName (mainDisplay, baseWin, POPUP_NAME);

/* XSetTransientForHint (mainDisplay, baseWin, mainWindow); */

   XSelectInput (mainDisplay, baseWin, KeyPressMask |
         ButtonPressMask | ExposureMask | StructureNotifyMask);
   XSelectInput (mainDisplay, dspWin,
         KeyPressMask | ButtonPressMask | ExposureMask);
   XSelectInput (mainDisplay, scrollWin,
         KeyPressMask | ButtonPressMask | ExposureMask);

   extraWinIndex = AddExtraWinInfo (baseWin, TRUE, FALSE,
         ChineseWinExposeHandler, ChineseWinEvHandler, CleanUpChinese);
   AddExtraWinInfo (dspWin, TRUE, FALSE, ChineseWinExposeHandler,
         ChineseWinEvHandler, CleanUpChinese);
   AddExtraWinInfo (scrollWin, TRUE, FALSE, ChineseWinExposeHandler,
         ChineseWinEvHandler, CleanUpChinese);

   chineseInit = TRUE;
   chineseWinExposed = FALSE;

   XMapWindow (mainDisplay, baseWin);
   XMapWindow (mainDisplay, dspWin);
   XMapWindow (mainDisplay, scrollWin);
}

static
void SelectChineseChar ()
{
   TieLooseEnds ();
   if (topSel != NULL) { HighLightReverse (); RemoveAllSel (); }

   justFetched = FALSE;
   lastSelected = NULL;

   spellIndex = 0;
   howSpelled = SPELL_CHINESE;

   DelAllEntries ();

   if (!chineseInit)
      InitChinese ();
   else
   {
      if (!extraWinInfo[extraWinIndex].mapped)
      {
         XMapWindow (mainDisplay, baseWin);
         extraWinInfo[extraWinIndex].mapped = FALSE;
      }
      if (chineseWinExposed)
      {
         RedrawBaseWin ();
         RedrawScrollWin ();
         RedrawDspWindow ();
      }
   }

   XWarpPointer (mainDisplay, None, baseWin, 0, 0, 0, 0,
         (int)(dspWinX+dspWinW/2), (int)(dspWinY+dspWinH/2));

   savedSelected = NULL;

   XSync (mainDisplay, False);
   Msg ("");
}

/* browse related variables */

#define DEF_BROWSE_SIZE 24

static int	defBrowseX = HALF_INCH, defBrowseY = HALF_INCH;
static int	curBrowseX, curBrowseY, curBrowseH = 0;
static int	maxBrowseWidth = (75*PIX_PER_INCH)/10;

static
void XBmToObjLoad (FileName)
   char	* FileName;
{
   char			full_name[MAXPATHLENGTH], msg[MAXSTRING];
   unsigned int		w, h;
   int			rc, x_hot, y_hot, len;
   Pixmap		bitmap;
   struct ObjRec	* obj_ptr;

   TieLooseEnds ();
   if (topSel != NULL) { HighLightReverse (); RemoveAllSel (); }

   len = strlen (FileName);
   if (len >= 4 && strcmp (&FileName[len-4], ".xbm") == 0)
      strcpy (full_name, FileName);
   else
      sprintf (full_name, "%s.xbm", FileName);

   rc = XReadBitmapFile (mainDisplay, mainWindow, full_name, &w, &h, &bitmap,
      &x_hot, &y_hot);

   if (rc != BitmapSuccess)
   {
      printf ("Can not import XBitmap file '%s'.\n", full_name);
      return;
   }

   obj_ptr = CreateXBmObj (w, h, w, h, bitmap, NULL);

   if (curBrowseX+w > maxBrowseWidth)
   {
      curBrowseX = 0;
      curBrowseY += curBrowseH;
      curBrowseH = 0;
   }
   if (curBrowseX+w > maxBrowseWidth)
   {  /* still bigger? */
      sprintf (msg, "Object width (%1d) > maxBrowseWidth (%1d)!",
            w, maxBrowseWidth);
      Msg (msg);
      printf ("%s\n", msg);
   }
   if (h > curBrowseH) curBrowseH = h;

   AddObj (NULL, topObj, obj_ptr);
   MoveObj (topObj, curBrowseX+defBrowseX, curBrowseY+defBrowseY);
   curBrowseX += w;
   DrawObj (drawWindow, topObj);
   RecordNewObjCmd ();
   SetFileModified (TRUE);
   justDupped = FALSE;
   XSync (mainDisplay, False);
}

static
void ReadInBitmaps ()
{
   char	inbuf[MAXSTRING+1];
   int	len;

   printf ("XBitmap filename to load> ");
   while (fgets (inbuf, MAXSTRING, stdin) != NULL)
   {
      len = strlen (inbuf);
      if (len > 0)
      {
         if (inbuf[--len] == '\n') inbuf[len] = '\0';
         XBmToObjLoad (inbuf);
      }
      printf ("\nXBitmap filename to load> ");
   }
   printf ("\n");
   rewind (stdin);
}

int main (argc, argv)
   int	argc;
   char	* argv[];
{
   char	s[MAXSTRING+1];
   char	op_name[MAXSTRING+1], file_name[MAXSTRING+1];
   char	* sp[6], * func_strp;
   int	i, quick_mode=FALSE;

   *file_name = '\0';
   for (i = 1; i < argc; i++)
   {
      if (strcmp (argv[i], "-display") == 0)
      {
         if (++i == argc) { printf ("Missing display name.\n"); exit (-1); }
         displayName = argv[i];
      }
      else if (strcmp (argv[i], "-bw") == 0)
         cmdLineBW = TRUE;
      else if (strcmp (argv[i], "-rv") == 0)
         cmdLineRV = TRUE;
      else if (strcmp (argv[i], "-nv") == 0)
         cmdLineRV = FALSE;
      else if (strcmp (argv[i], "-fg") == 0)
      {
         if (++i == argc)
         { fprintf (stderr, "Missing foreground color.\n"); exit (-1); }
         cmdLineForeground = (char *) calloc (80, sizeof(char));
         strcpy (cmdLineForeground, argv[i]);
      }
      else if (strcmp (argv[i], "-bg") == 0)
      {
         if (++i == argc)
         { fprintf (stderr, "Missing background color.\n"); exit (-1); }
         cmdLineBackground = (char *) calloc (80, sizeof(char));
         strcpy (cmdLineBackground, argv[i]);
      }
      else if (strcmp (argv[i], "-bd") == 0)
      {
         if (++i == argc)
         { fprintf (stderr, "Missing border color.\n"); exit (-1); }
         cmdLineBorder = (char *) calloc (80, sizeof(char));
         strcpy (cmdLineBorder, argv[i]);
      }
      else if (strcmp (argv[i], "-geometry") == 0)
      {
         if (++i == argc) { printf ("Missing geometry spec.\n"); exit (-1); }
         strcat (geometrySpec, argv[i]);
         geometrySpecified = TRUE;
      }
      else if (*argv[i] == '=')
      {
         strcpy (geometrySpec, argv[i]);
         geometrySpecified = TRUE;
      }
      else if (*argv[i] == '-')
      {
         if (strncmp (argv[i], "-w", 2) == 0)
            maxBrowseWidth = atoi(&(argv[i][2]));
         else if (strncmp (argv[i], "-x", 2) == 0)
            defBrowseX = atoi(&(argv[i][2]));
         else if (strncmp (argv[i], "-y", 2) == 0)
            defBrowseY = atoi(&(argv[i][2]));
         else if (strcmp (argv[i], "-quick") == 0)
            quick_mode = TRUE;
         else
         {
            printf("Usage: ctgif [-display display] [=geometry] %s\n",
                  "[-w#] [-x#] [-y]");
            exit(-1);
         }
      }
      else
         strcpy (file_name, argv[i]);
   }

   curBrowseX = 0;
   curBrowseY = 0;
   curBrowseH = 0;

/* malloc_debug (1); */

   if (quick_mode)
   {
      if (*file_name == '\0')
         MainLoop ("justinit", "", &func_strp,
               &sp[0], &sp[1], &sp[2], &sp[3], &sp[4], &sp[5]);
      else
         MainLoop ("justinit", file_name, &func_strp,
               &sp[0], &sp[1], &sp[2], &sp[3], &sp[4], &sp[5]);
      DeallocStrings (&func_strp,&sp[0],&sp[1],&sp[2],&sp[3],&sp[4],&sp[5]);
      ReadInBitmaps ();
      MainLoop ("save", file_name, &func_strp,
            &sp[0], &sp[1], &sp[2], &sp[3], &sp[4], &sp[5]);
      DeallocStrings (&func_strp,&sp[0],&sp[1],&sp[2],&sp[3],&sp[4],&sp[5]);
      MainLoop ("quit", file_name, &func_strp,
            &sp[0], &sp[1], &sp[2], &sp[3], &sp[4], &sp[5]);
      DeallocStrings (&func_strp,&sp[0],&sp[1],&sp[2],&sp[3],&sp[4],&sp[5]);
      return (0);
   }

   if (*file_name == '\0')
      MainLoop ("init", "", &func_strp,
            &sp[0], &sp[1], &sp[2], &sp[3], &sp[4], &sp[5]);
   else
      MainLoop ("init", file_name, &func_strp,
            &sp[0], &sp[1], &sp[2], &sp[3], &sp[4], &sp[5]);

   while (TRUE)
   {
      strcpy (s, func_strp);
      s[4] = '\0';

      DeallocStrings (&func_strp,&sp[0],&sp[1],&sp[2],&sp[3],&sp[4],&sp[5]);

      if (strcmp (s, "Quit") == 0)
      {
         MainLoop ("quit", file_name, &func_strp,
               &sp[0], &sp[1], &sp[2], &sp[3], &sp[4], &sp[5]);
         DeallocStrings (&func_strp,&sp[0],&sp[1],&sp[2],&sp[3],&sp[4],&sp[5]);
         break;
      }
      else if (strcmp(s,"Prob") == 0)
         ReadInBitmaps ();
      else if (strcmp(s,"Solv") == 0)
         SelectChineseChar ();
      else if (strcmp(s,"Anim") == 0)
         SearchAgain ();
      else if (strcmp(s,"Esca") == 0)
         TsaiHsing ();
      else if (strcmp(s,"Simu") == 0)
         TsaiHsing ();

      Msg ("Returned from basic driver.");
      MainLoop (op_name, file_name, &func_strp,
            &sp[0], &sp[1], &sp[2], &sp[3], &sp[4], &sp[5]);
   }
   return (0);
}
