/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, 1991, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /amnt/kona/tangram/u/william/X11/TGIF2/Chinese/RCS/tsaihsing.c,v 1.20 1992/12/14 04:49:08 william Exp $";
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include "../const.h"
#include "../types.h"

#include "../attr.e"
#include "../choice.e"
#include "../cmd.e"
#include "../cursor.e"
#include "../dialog.e"
#include "../drawing.e"
#include "../dup.e"
#include "../file.e"
#include "../grid.e"
#include "../mark.e"
#include "../move.e"
#include "../msg.e"
#include "../names.e"
#include "../obj.e"
#include "../select.e"
#include "../setup.e"

extern char	* getenv ();

static char defAddrFile[] = "/u/tangram/u/william/TsaiHsing/Addresses";
static char defTmp1File[] = "/u/tangram/u/william/TsaiHsing/template_1.obj";
static char defTmp2File[] = "/u/tangram/u/william/TsaiHsing/template_2.obj";

static char address_name[MAXSTRING];
static char template_1_name[MAXSTRING];
static char template_2_name[MAXSTRING];

static struct ObjRec	* firstPageTopObj=NULL, * firstPageBotObj=NULL;
static struct ObjRec	* firstPageTopAnchor=NULL, * firstPageBotAnchor=NULL;
static struct ObjRec	* secondPageTopObj=NULL, * secondPageBotObj=NULL;
static struct ObjRec	* secondPageTopAnchor=NULL, * secondPageBotAnchor=NULL;

static FILE	* fp, * fp1, * fp2;

struct HashRec {
   char			key1[80], key2[80], key3[80];
   struct ObjRec	* value;
   struct HashRec	* next, * prev;
};

static struct HashRec	* * hashTopArray = NULL;
static struct HashRec	* * hashBotArray = NULL;

static char	defKey1[] = "Name=";
static char	defKey2[] = "Address=";
static char	defKey3[] = "City=";

static char	firstKey[MAXSTRING];
static char	secondKey[MAXSTRING];
static char	thirdKey[MAXSTRING];

static int	numStateConstraints=0;
static int	* lenStateConstraints=NULL;
static char	* * stateConstraints=NULL;

static
void SelectEverything ()
{
   register struct ObjRec	* obj_ptr;
   register struct SelRec	* sel_ptr;

   topSel = botSel = NULL;
   for (obj_ptr=botObj; obj_ptr !=NULL; obj_ptr=obj_ptr->prev)
   {
      sel_ptr = (struct SelRec *) calloc (1, sizeof(struct SelRec));
      sel_ptr->next = topSel;
      sel_ptr->obj = obj_ptr;
      sel_ptr->prev = NULL;
      if (topSel == NULL)
         botSel = sel_ptr;
      else
         topSel->prev = sel_ptr;
      topSel = sel_ptr;
   }
   UpdSelBBox ();
}

static int	gotXDefaults=FALSE;

static
void GetXDefaults ()
{
   char	* c_ptr;

   if (gotXDefaults) return;

   if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "AddrFile")) != NULL)
      strcpy (address_name, c_ptr);
   else
      strcpy (address_name, defAddrFile);
   if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "Tmp1File")) != NULL)
      strcpy (template_1_name, c_ptr);
   else
      strcpy (template_1_name, defTmp1File);
   if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "Tmp2File")) != NULL)
      strcpy (template_2_name, c_ptr);
   else
      strcpy (template_2_name, defTmp2File);
   if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "FirstKey")) != NULL)
      strcpy (firstKey, c_ptr);
   else
      strcpy (firstKey, defKey1);
   if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "SecondKey")) != NULL)
      strcpy (secondKey, c_ptr);
   else
      strcpy (secondKey, defKey2);
   if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "ThirdKey")) != NULL)
      strcpy (thirdKey, c_ptr);
   else
      strcpy (thirdKey, defKey3);

   numStateConstraints = 0;
   if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "StateConstraints"))!=NULL)
   {
      int	len=strlen(c_ptr), count=0;
      char	* buf, * saved_c_ptr;

      saved_c_ptr = c_ptr;

      /* count the number of fields first */
      while ((*c_ptr == ' ' || *c_ptr == '\t') && *c_ptr != '\0') c_ptr++;
      while (*c_ptr != '\0')
      {
         while (*c_ptr != '\n' && *c_ptr != '\0') c_ptr++;
         if (c_ptr != '\0') numStateConstraints++;
         while ((*c_ptr==' ' || *c_ptr=='\t' || *c_ptr=='\n') && *c_ptr!='\0')
            c_ptr++;
      }

      c_ptr = saved_c_ptr;
      stateConstraints = (char * *) calloc (numStateConstraints,sizeof(char *));
      lenStateConstraints = (int *) calloc (numStateConstraints,sizeof(int *));

      while ((*c_ptr == ' ' || *c_ptr == '\t') && *c_ptr != '\0') c_ptr++;
      while (*c_ptr != '\0')
      {
         buf = stateConstraints[count] = (char *) calloc (len+1, sizeof(char));
         while (*c_ptr != '\n' && *c_ptr != '\0') *buf++ = *c_ptr++;
         *buf = '\0';
         lenStateConstraints[count] = strlen (stateConstraints[count]);

         if (c_ptr != '\0') count++;
         while ((*c_ptr==' ' || *c_ptr=='\t' || *c_ptr=='\n') && *c_ptr!='\0')
            c_ptr++;
      }
      if (numStateConstraints == 0)
         fprintf (stderr, "No StateConstraints.\n");
      else
      {
         int	first=TRUE;

         fprintf (stderr, "\nStateConstraints are:\n       ");
         for (count = 0; count < numStateConstraints; count++)
         {
            if (strcmp(stateConstraints[count], "||") == 0)
            {
               fprintf (stderr, "\n     %s", stateConstraints[count]);
               first = TRUE;
            }
            else
            {
               if (first)
               {
                  fprintf (stderr, " '%s'", stateConstraints[count]);
                  first = FALSE;
               }
               else
                  fprintf (stderr, " && '%s'", stateConstraints[count]);
            }
         }
         fprintf (stderr, "\n");
      }
   }
   gotXDefaults = TRUE;
}

static
struct ObjRec * DesiredObj (obj_ptr)
   struct ObjRec	* obj_ptr;
{
   register struct ObjRec	* ptr;
   struct ObjRec		* return_ptr;

   if (obj_ptr->type == OBJ_GROUP)
   {
      if (obj_ptr->fattr != NULL) return (obj_ptr);

      for (ptr=obj_ptr->detail.r->last; ptr!=NULL; ptr=ptr->prev)
         if ((return_ptr = DesiredObj (ptr)) != NULL)
            return (return_ptr);
   }

   return (NULL);
}

static
int ProcessPage (PageNum)
   int	PageNum;
{
   struct ObjRec	* obj_ptr, * * top_anchor=NULL, * * bot_anchor=NULL;

   switch (PageNum)
   {
      case 1:
         firstPageTopObj = topObj; firstPageBotObj = botObj;
         firstPageTopAnchor = NULL; firstPageBotAnchor = NULL;
         top_anchor = &firstPageTopAnchor; bot_anchor = &firstPageBotAnchor;
         break;
      case 2:
         secondPageTopObj = topObj; secondPageBotObj = botObj;
         secondPageTopAnchor = NULL; secondPageBotAnchor = NULL;
         top_anchor = &secondPageTopAnchor; bot_anchor = &secondPageBotAnchor;
         break;
   }

   for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      if (DesiredObj (obj_ptr) != NULL)
      {
         if (*top_anchor == NULL)
            *top_anchor = obj_ptr;
         else if ((*top_anchor)->obbox.lty < obj_ptr->obbox.lty)
         {
            *bot_anchor = obj_ptr;
            break;
         }
         else
         {
            *bot_anchor = *top_anchor;
            *top_anchor = obj_ptr;
            break;
         }
      }
   }
   if (*top_anchor==NULL || *bot_anchor==NULL)
   {
      printf ("Can not find all the anchors in template_%1d\n", PageNum);
      return (FALSE);
   }

   topObj = botObj = NULL;
   return (TRUE);
}

static
int LoadTemplate ()
{
   char			msg[MAXPATHLENGTH], tmp_filename[MAXPATHLENGTH];
   int			read_status, tmp_linenum;
   struct ObjRec	* obj_ptr;

   if ((fp1 = fopen (template_1_name, "r")) == NULL)
   {
      printf ("Can not open '%s'.  Abort!\n", template_1_name);
      return (FALSE);
   }
   if ((fp2 = fopen (template_2_name, "r")) == NULL)
   {
      fclose (fp1);
      printf ("Can not open '%s'.  Abort!\n", template_2_name);
      return (FALSE);
   }

   TieLooseEnds ();
   CleanUpDrawingWindow ();
   SetFileModified (FALSE);

   strcpy (tmp_filename, scanFileName);
   tmp_linenum = scanLineNum;
   strcpy (scanFileName, template_2_name);
   scanLineNum = 0;

   sprintf (msg, "Loading '%s' ...", template_2_name);
   Msg (msg);
   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   XClearWindow (mainDisplay, drawWindow);
   somethingHighLighted = FALSE;

   while ((read_status = ReadObj (fp2, &obj_ptr)) == TRUE)
      if (obj_ptr != NULL)
      {
         AdjForOldVersion (obj_ptr);
         AddObj (NULL, topObj, obj_ptr);
         if (PointInBBox (obj_ptr->x, obj_ptr->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr->bbox, drawWinBBox))
            DrawObj (drawWindow, obj_ptr);
      }

   strcpy (scanFileName, tmp_filename);
   scanLineNum = tmp_linenum;

   if (read_status == INVALID)
   {
      sprintf (msg, "File version too large (=%1d).  Load aborted!",
            fileVersion);
      Msg (msg);
      DrawPaperBoundary ();
      RedrawGridLines ();
      SetDefaultCursor (mainWindow);
      SetDefaultCursor (drawWindow);
      return (FALSE);
   }
   fclose (fp2);

   RedrawTitleWindow ();
   justDupped = FALSE;
   SetDefaultCursor (mainWindow);
   SetDefaultCursor (drawWindow);

   if (!ProcessPage (2)) return (FALSE);

   CleanUpDrawingWindow ();
   SetFileModified (FALSE);

   strcpy (tmp_filename, scanFileName);
   tmp_linenum = scanLineNum;
   strcpy (scanFileName, template_1_name);
   scanLineNum = 0;

   sprintf (msg, "Loading '%s' ...", template_1_name);
   Msg (msg);
   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   XClearWindow (mainDisplay, drawWindow);
   somethingHighLighted = FALSE;

   while ((read_status = ReadObj (fp1, &obj_ptr)) == TRUE)
      if (obj_ptr != NULL)
      {
         AdjForOldVersion (obj_ptr);
         AddObj (NULL, topObj, obj_ptr);
         if (PointInBBox (obj_ptr->x, obj_ptr->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr->bbox, drawWinBBox))
            DrawObj (drawWindow, obj_ptr);
      }

   strcpy (scanFileName, tmp_filename);
   scanLineNum = tmp_linenum;

   if (read_status == INVALID)
   {
      sprintf (msg, "File version too large (=%1d).  Load aborted!",
            fileVersion);
      Msg (msg);
      DrawPaperBoundary ();
      RedrawGridLines ();
      SetDefaultCursor (mainWindow);
      SetDefaultCursor (drawWindow);
      return (FALSE);
   }
   fclose (fp1);

   RedrawTitleWindow ();
   justDupped = FALSE;
   SetDefaultCursor (mainWindow);
   SetDefaultCursor (drawWindow);

   if (!ProcessPage (1)) return (FALSE);

   CleanUpDrawingWindow ();
   ClearAndRedrawDrawWindow ();

   return (TRUE);
}

static
char * ReadLine ()
{
   char	* line, * c_ptr;
   int	unit_size = 8;
   int	len, cur_size=unit_size;

   line = (char *) calloc (cur_size+1, sizeof (char));
   if (fgets (line, unit_size, fp) == NULL)
   {
      cfree (line);
      return (NULL);
   }

   c_ptr = line;
   len = strlen (line);
   while (line[len-1] != '\r' && line[len-1] != '\n')
   {
      cur_size += unit_size-1;
      line = (char *) realloc (line, sizeof(char)*cur_size+1);
      c_ptr = &(line[len]);

      if (fgets (c_ptr, unit_size, fp) == NULL)
      {
         cfree (line);
         return (NULL);
      }
      len += strlen (c_ptr);
   }
   line[len-1] = '\0';
   return (line);
}

#define LASTNAME 1
#define FIRSTNAME 2
#define MIDDLENAME 3
#define ADDRESS 4
#define CITY 6
#define STATEZIP 7
#define HOMEPHONE 8
#define WORKPHONE 9
#define E 10
#define J 11
#define S 12
#define OCCUPATION 13
#define SPOUSENAME 15
#define SPOUSEWORKPHONE 16
#define SPOUSEOCCUPATION 17

static int	entryY, incY, fieldCount=0;
static char	* * fieldName=NULL, * * fieldContent=NULL;

static
void TrimmedStrCopy (to_str, from_str)
   char	* to_str, * from_str;
{
   register char	* left_ptr, * right_ptr;
   register int		len;

   *to_str = '\0';
   if (*from_str == '\0') return;

   len = strlen (from_str)-1;
   right_ptr = &from_str[len];
   while (*right_ptr == ' ')
      if (right_ptr == from_str)
         return;
      else
         right_ptr--;

   left_ptr = from_str;
   while (*left_ptr == ' ')
      if (left_ptr == right_ptr)
         return;
      else
         left_ptr++;

   while (*left_ptr == '"' && *right_ptr == '"')
      if (left_ptr == right_ptr)
         return;
      else
      {
         left_ptr++;
         right_ptr--;
      }

   right_ptr[1] = '\0';
   strcpy (to_str, left_ptr);
}

static
void ParseFirstLine (line)
   char	* line;
{
   register int		i;
   register char	* c_ptr, * buf_ptr;
   char			buf[MAXSTRING];

   if (fieldName != NULL)
      for (i = 0; i < fieldCount; i++)
      {
         cfree (fieldName[i]);
         cfree (fieldContent[i]);
      }
   cfree (fieldName);
   cfree (fieldContent);
   fieldName = NULL;
   fieldContent = NULL;
   fieldCount = 0;

   for (c_ptr = line; *c_ptr != '\0'; c_ptr++)
      if (*c_ptr == '\t')
         fieldCount++;

   fieldName = (char * *) calloc (fieldCount, sizeof (char *));
   fieldContent = (char * *) calloc (fieldCount, sizeof (char *));

   for (c_ptr=line, i=0, buf_ptr=buf; *c_ptr != '\0'; c_ptr++)
      if (*c_ptr == '\t')
      {
         *buf_ptr = '\0';
         fieldName[i] = (char *) calloc (strlen (buf)+1, sizeof (char));
         TrimmedStrCopy (fieldName[i], buf);
         fieldContent[i] = (char *) calloc (MAXSTRING+1, sizeof (char));
         i++;
         buf_ptr = buf;
      }
      else
         *buf_ptr++ = *c_ptr;
}

static
void ParseOtherLine (line)
   char	* line;
{
   register int		i;
   register char	* c_ptr, * buf_ptr;
   char			buf[MAXSTRING];

   for (c_ptr=line, i=0, buf_ptr=buf; *c_ptr != '\0'; c_ptr++)
      if (*c_ptr == '\t')
      {
         *buf_ptr = '\0';
         TrimmedStrCopy (fieldContent[i], buf);
         i++;
         buf_ptr = buf;
      }
      else
         *buf_ptr++ = *c_ptr;
}

static
int ReadFirstLine ()
{
   char	* line;

   entryY = firstPageTopAnchor->obbox.lty;
   incY = firstPageTopAnchor->obbox.rby - firstPageTopAnchor->obbox.lty;

   if ((line = ReadLine()) == NULL) return (FALSE);

   ParseFirstLine (line);
   cfree (line);
}

static
void UpdateSingleField (obj_ptr, field_name, field_value)
   struct ObjRec	* obj_ptr;
   char			* field_name, * field_value;
{
   register struct AttrRec	* attr_ptr;

   for (attr_ptr=obj_ptr->lattr; attr_ptr!=NULL; attr_ptr=attr_ptr->prev)
      if (strcmp (attr_ptr->name, field_name) == 0)
      {
         strcpy (attr_ptr->s, field_value);
         UpdAttr (attr_ptr);
         return;
      }
   printf ("Warning:  can not find field '%s'\n", field_name);
}

static
void ConstructName (name)
   char	* name;
{
   *name = '\0';
   if (*fieldContent[LASTNAME] == '\0') return;
   if (*fieldContent[FIRSTNAME] == '\0')
   {  /* no first name */
      if (*fieldContent[MIDDLENAME] == '\0')
         strcpy (name, fieldContent[LASTNAME]);
      else
         sprintf (name, "%s, %s", fieldContent[LASTNAME],
               fieldContent[MIDDLENAME]);
   }
   else
   {  /* has first name */
      if (*fieldContent[MIDDLENAME] == '\0')
         sprintf (name, "%s, %s", fieldContent[LASTNAME],
               fieldContent[FIRSTNAME]);
      else
         sprintf (name, "%s, %s %s", fieldContent[LASTNAME],
               fieldContent[FIRSTNAME], fieldContent[MIDDLENAME]);
   }
   if (*name != '\0' && (*name < 'A' || *name > 'Z'))
   {
      if (*name >= 'a' && *name <= 'z')
      {
         printf ("Converting '%c%s' to '%c%s'.\n", *name, name+1,
               *name-'a'+'A', name+1);
         *name = *name-'a'+'A';
      }
      else
         printf ("Warning:  Probably illegal name '%s'.\n", name);
   }
}

static
void UpdateFields (obj_ptr, has_spouse_info)
   struct ObjRec	* obj_ptr;
   int			has_spouse_info;
{
   char	name[MAXSTRING];

   ConstructName (name);
   UpdateSingleField (obj_ptr, "Name=", name);
   UpdateSingleField (obj_ptr, "Address=", fieldContent[ADDRESS]);
   UpdateSingleField (obj_ptr, "City=", fieldContent[CITY]);
   UpdateSingleField (obj_ptr, "StateZip=", fieldContent[STATEZIP]);
   UpdateSingleField (obj_ptr, "HomePhone=", fieldContent[HOMEPHONE]);
   UpdateSingleField (obj_ptr, "WorkPhone=", fieldContent[WORKPHONE]);
   UpdateSingleField (obj_ptr, "E=", fieldContent[E]);
   UpdateSingleField (obj_ptr, "J=", fieldContent[J]);
   UpdateSingleField (obj_ptr, "S=", fieldContent[S]);
   UpdateSingleField (obj_ptr, "Occupation=", fieldContent[OCCUPATION]);
   if (has_spouse_info)
   {
      UpdateSingleField (obj_ptr, "SpouseName=", fieldContent[SPOUSENAME]);
      UpdateSingleField (obj_ptr, "SpouseWorkPhone=",
            fieldContent[SPOUSEWORKPHONE]);
      UpdateSingleField (obj_ptr, "SpouseOccupation=",
            fieldContent[SPOUSEOCCUPATION]);
   }
}

static
int HasSpouseInformation ()
{
   return (*fieldContent[SPOUSENAME]!='\0' ||
         *fieldContent[SPOUSEWORKPHONE]!='\0' ||
         *fieldContent[SPOUSEOCCUPATION]!='\0');
}

static
void ReadRest ()
{
   char			* line;
   int			dy=0, line_count=0, has_spouse_info;
   int			cur_y = firstPageTopAnchor->obbox.lty;
   struct ObjRec	* obj_ptr;

   while ((line = ReadLine()) != NULL)
   {
      ParseOtherLine (line);
      if (strcmp (fieldContent[LASTNAME], "ZZZ") == 0)
      {
         printf ("Normal EOF detected with %1d lines read.\n", line_count);
         cfree (line);
         return;
      }

      if (has_spouse_info = HasSpouseInformation ())
      {
         obj_ptr = DupObj (firstPageBotAnchor);
         MoveObj (obj_ptr, firstPageTopAnchor->obbox.ltx-obj_ptr->obbox.ltx,
               cur_y-firstPageBotAnchor->obbox.lty);
      }
      else
      {
         obj_ptr = DupObj (firstPageTopAnchor);
         MoveObj (obj_ptr, firstPageTopAnchor->obbox.ltx-obj_ptr->obbox.ltx,
               cur_y-firstPageTopAnchor->obbox.lty);
      }
      UpdateFields (obj_ptr, has_spouse_info);
      AdjObjSplineVs (obj_ptr);
      AddObj (NULL, topObj, obj_ptr);
      if (PointInBBox (obj_ptr->x, obj_ptr->y, drawWinBBox) ||
            BBoxIntersect (obj_ptr->bbox, drawWinBBox))
         DrawObj (drawWindow, obj_ptr);

      cur_y += (has_spouse_info) ? 2*incY : incY;
      line_count++;
      cfree (line);
   }
   printf ("Warning:  EOF not detected with %1d lines read!\n", line_count);
}

static
int LoadTsaiHsing ()
{
   struct ObjRec	* obj_ptr, * obj_ptr_1;

   if (!LoadTemplate ()) return (FALSE);

   if ((fp = fopen (address_name, "r")) == NULL)
   {
      fclose (fp2);
      fclose (fp1);
      printf ("Can not open '%s'.  Abort!\n", address_name);
      return (FALSE);
   }

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   ClearFileInfo ();
   for (obj_ptr = firstPageBotObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      if (obj_ptr != firstPageTopAnchor && obj_ptr != firstPageBotAnchor)
      {
         obj_ptr_1 = DupObj (obj_ptr);
         AdjObjSplineVs (obj_ptr_1);
         AddObj (NULL, topObj, obj_ptr_1);
         if (PointInBBox (obj_ptr_1->x, obj_ptr_1->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr_1->bbox, drawWinBBox))
            DrawObj (drawWindow, obj_ptr_1);
      }
   }
   SetFileModified (TRUE);

   if (!ReadFirstLine ())
   {
      SetDefaultCursor (mainWindow);
      SetDefaultCursor (drawWindow);
      return (FALSE);
   }
   ReadRest ();
   fclose (fp);

   SetDefaultCursor (mainWindow);
   SetDefaultCursor (drawWindow);

   CleanUpCmds ();

   return (TRUE);
}

static
void ReadTsaiHsing ()
{
   int	do_not_save=FALSE;

   while (fileModified)
   {
      switch (YesNoCancel ("File modified, save file before open? [ync](y)"))
      {
         case CONFIRM_YES: SaveFile (); break;
         case CONFIRM_NO: do_not_save = TRUE; SetFileModified (FALSE); break;
         case CONFIRM_CANCEL: return;
      }
   }
   GetXDefaults ();
   if (!LoadTsaiHsing () && do_not_save) SetFileModified (TRUE);

   RedrawTitleWindow ();
   SetDefaultCursor (mainWindow);
   if (curChoice == DRAWTEXT)
      SetTextCursor (drawWindow);
   else
      SetDefaultCursor (drawWindow);
}

static
void FilterFields (template, obj_ptr)
   struct ObjRec	* template, * obj_ptr;
{
   register struct AttrRec	* tmp_attr, * attr_ptr;

   for (attr_ptr=obj_ptr->lattr; attr_ptr!=NULL; attr_ptr=attr_ptr->prev)
      for (tmp_attr=template->lattr; tmp_attr!=NULL; tmp_attr=tmp_attr->prev)
         if (strcmp (attr_ptr->name, tmp_attr->name) == 0)
         {
            strcpy (tmp_attr->s, attr_ptr->s);
            UpdAttr (tmp_attr);
            break;
         }
}

static
void DupPartialGroupObj (GroupPtr, new_obj, obj_with_keys, template)
   struct GroupRec	* GroupPtr;
   struct ObjRec	* new_obj, * obj_with_keys, * template;
{
   struct GroupRec	* group_ptr;
   struct ObjRec	* top_obj, * bot_obj;
   struct ObjRec	* from_obj_ptr, * to_obj_ptr;

   group_ptr = (struct GroupRec *) calloc (1, sizeof(struct GroupRec));
   top_obj = bot_obj = NULL;
   from_obj_ptr = GroupPtr->last;
   for ( ; from_obj_ptr != NULL; from_obj_ptr = from_obj_ptr->prev)
   {
      if (from_obj_ptr == obj_with_keys)
         to_obj_ptr = template;
      else
      {
         if (from_obj_ptr->type == OBJ_GROUP)
         {
            to_obj_ptr = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));
            DupObjBasics (from_obj_ptr, to_obj_ptr);
            DupPartialGroupObj (from_obj_ptr->detail.r, to_obj_ptr,
                  obj_with_keys, template);
            DupAttrs (from_obj_ptr, to_obj_ptr);
         }
         else
            to_obj_ptr = DupObj (from_obj_ptr);
         AdjObjSplineVs (to_obj_ptr);
      }
      to_obj_ptr->next = top_obj;
      if (top_obj == NULL)
         group_ptr->last = to_obj_ptr;
      else
         top_obj->prev = to_obj_ptr;
      top_obj = to_obj_ptr;
   }
   top_obj->prev = NULL;
   group_ptr->first = top_obj;

   new_obj->detail.r = group_ptr;
}

#define ONE_KIND (TRUE)
#define TWO_KINDS (FALSE)

static
int FilterTsaiHsing (one_kind)
   int	one_kind;
{
   int			cur_y;
   int			h, top_anchor_h, bot_anchor_h, bad_count, good_count;
   struct ObjRec	* obj_ptr, * obj_ptr_1, * template;
   struct ObjRec	* saved_top_obj, * saved_bot_obj, * obj_with_keys;
   struct ObjRec	* new_obj;

   saved_top_obj = topObj;
   saved_bot_obj = botObj;
   topObj = botObj = NULL;

   ClearAndRedrawDrawWindow ();

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   for (obj_ptr = firstPageBotObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      if (obj_ptr != firstPageTopAnchor && obj_ptr != firstPageBotAnchor)
      {
         obj_ptr_1 = DupObj (obj_ptr);
         AdjObjSplineVs (obj_ptr_1);
         AddObj (NULL, topObj, obj_ptr_1);
         if (PointInBBox (obj_ptr_1->x, obj_ptr_1->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr_1->bbox, drawWinBBox))
            DrawObj (drawWindow, obj_ptr_1);
      }
   }
   SetFileModified (TRUE);
   RedrawTitleWindow ();

   bad_count = good_count = 0;

   top_anchor_h = firstPageTopAnchor->obbox.rby - firstPageTopAnchor->obbox.lty;
   bot_anchor_h = firstPageBotAnchor->obbox.rby - firstPageBotAnchor->obbox.lty;
   cur_y = firstPageTopAnchor->obbox.lty;
   for (obj_ptr = saved_bot_obj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      if ((obj_with_keys = DesiredObj (obj_ptr)) != NULL)
      {
         h = obj_ptr->obbox.rby - obj_ptr->obbox.lty;
         if (one_kind)
         {
            template = DupObj (firstPageTopAnchor);
            MoveObj (template,
                  firstPageTopAnchor->obbox.ltx-template->obbox.ltx,
                  cur_y-firstPageTopAnchor->obbox.lty);
         }
         else
         {
            if (h == top_anchor_h)
            {
               template = DupObj (firstPageTopAnchor);
               MoveObj (template,
                     firstPageTopAnchor->obbox.ltx-template->obbox.ltx,
                     cur_y-firstPageTopAnchor->obbox.lty);
            }
            else if (h == bot_anchor_h)
            {
               template = DupObj (firstPageBotAnchor);
               MoveObj (template,
                     firstPageTopAnchor->obbox.ltx-template->obbox.ltx,
                     cur_y-firstPageBotAnchor->obbox.lty);
            }
            else
            {
               bad_count++;
               printf ("Can not find macehs object height %1d.  Skipped!\n", h);
               continue;
            }
         }
         FilterFields (template, obj_with_keys);
         if (obj_ptr == obj_with_keys)
            new_obj = template;
         else
         {
            new_obj = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));
            DupObjBasics (obj_ptr, new_obj);
            DupPartialGroupObj (obj_ptr->detail.r, new_obj,
                  obj_with_keys, template);
            DupAttrs (obj_ptr, new_obj);
         }
         good_count++;
         AdjObjSplineVs (new_obj);
         AddObj (NULL, topObj, new_obj);
         if (PointInBBox (new_obj->x, new_obj->y, drawWinBBox) ||
               BBoxIntersect (new_obj->bbox, drawWinBBox))
            DrawObj (drawWindow, new_obj);

         cur_y += h;
      }
      FreeObj (obj_ptr);
   }

   SetDefaultCursor (mainWindow);
   SetDefaultCursor (drawWindow);

   if (bad_count != 0)
      printf ("Total of %1d objects have bad heights.\n", bad_count);
   printf ("%1d objects filtered.\n", good_count);
   return (TRUE);
}

static
void FilterDrawing ()
{
   int	do_not_save=FALSE;

   TieLooseEnds ();
   SetCurChoice (NOTHING);
   if (topSel!=NULL) { HighLightReverse (); RemoveAllSel (); }

   GetXDefaults ();
   if (!ProcessTemplates ()) return;

   FilterTsaiHsing (TWO_KINDS);

   CleanUpCmds ();
}

static
void CleanUpHashTbl ()
{
   register int		i;
   struct HashRec	* hash_ptr;

   if (hashBotArray == NULL)
   {
      hashTopArray = (struct HashRec * *) calloc (27, sizeof(struct HashRec *));
      hashBotArray = (struct HashRec * *) calloc (27, sizeof(struct HashRec *));
      for (i = 0; i < 27; i++)
         hashTopArray[i] = hashBotArray[i] = NULL;
      return;
   }
   for (i = 0; i < 27; i++)
   {
      if (hashBotArray[i] != NULL)
      {
         for (hash_ptr = hashBotArray[i]; hash_ptr != NULL;
               hash_ptr = hash_ptr->prev)
            cfree (hash_ptr);
         hashTopArray[i] = hashBotArray[i] = NULL;
      }
   }
}

static
int ProcessTemplates ()
{
   int			saved_file_modified;
   struct ObjRec	* saved_top_obj, * saved_bot_obj;

   saved_top_obj = topObj;
   saved_bot_obj = botObj;
   saved_file_modified = fileModified;
   topObj = botObj = NULL;

   if (!LoadTemplate ())
   {
      fileModified = saved_file_modified;
      topObj = saved_top_obj;
      botObj = saved_bot_obj;
      return (FALSE);
   }
   fileModified = saved_file_modified;
   topObj = saved_top_obj;
   botObj = saved_bot_obj;
   ClearAndRedrawDrawWindow ();
   RedrawTitleWindow ();

   return (TRUE);
}

static
struct ObjRec	* FindNextObj (obj_ptr)
   struct ObjRec	* obj_ptr;
{
   for (obj_ptr = obj_ptr->prev; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
      if (DesiredObj (obj_ptr) != NULL)
         return (obj_ptr);
   return (NULL);
}

static char	dumpName[] = "Directory/Page";

static
void DumpTsaiHsing ()
{
   FILE			* page_fp;
   int			saved_file_modified, cur_y, page_num, h, count;
   int			saved_file_defined;
   int			short_name, total = 0;
   char			file_name[MAXPATHLENGTH], * rest;
   char			saved_file_name[MAXPATHLENGTH];
   char			saved_dir_name[MAXPATHLENGTH];
   char			saved_sym_dir[MAXPATHLENGTH];
   struct ObjRec	* saved_top_obj, * saved_bot_obj, * copy, * bot_anchor;
   struct ObjRec	* next_obj=NULL, * obj_ptr, * obj_ptr_1, * obj_ptr_2;

   saved_top_obj = topObj;
   saved_bot_obj = botObj;
   saved_file_modified = fileModified;
   saved_file_defined = curFileDefined;
   strcpy (saved_file_name, curFileName);
   strcpy (saved_dir_name, curDir);
   strcpy (saved_sym_dir, curSymDir);
   topObj = botObj = NULL;
   ClearAndRedrawDrawWindow ();

   ClearFileInfo ();
   for (obj_ptr = firstPageBotObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      if (obj_ptr != firstPageTopAnchor && obj_ptr != firstPageBotAnchor)
      {
         obj_ptr_1 = DupObj (obj_ptr);
         AdjObjSplineVs (obj_ptr_1);
         AddObj (NULL, topObj, obj_ptr_1);
         if (PointInBBox (obj_ptr_1->x, obj_ptr_1->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr_1->bbox, drawWinBBox))
            DrawObj (drawWindow, obj_ptr_1);
      }
   }
   SetFileModified (TRUE);
   RedrawTitleWindow ();

   count = 0;
   page_num = 1;
   cur_y = firstPageTopAnchor->obbox.lty;
   bot_anchor = firstPageBotAnchor;
   for (obj_ptr = saved_bot_obj; obj_ptr != NULL; obj_ptr = next_obj)
   {
      if (DesiredObj (obj_ptr) != NULL)
      {
         h = obj_ptr->obbox.rby-obj_ptr->obbox.lty;
         copy = DupObj (obj_ptr);
         MoveObj (copy, firstPageTopAnchor->obbox.ltx-copy->obbox.ltx,
               cur_y-obj_ptr->obbox.lty);
         AdjObjSplineVs (copy);
         AddObj (NULL, topObj, copy);
         count++;
         if (PointInBBox (copy->x, copy->y, drawWinBBox) ||
               BBoxIntersect (copy->bbox, drawWinBBox))
            DrawObj (drawWindow, copy);
         cur_y += h;
         next_obj = FindNextObj (obj_ptr);
         if (next_obj == NULL ||
               next_obj->obbox.rby-next_obj->next->obbox.lty+cur_y >
               bot_anchor->obbox.rby)
         {  /* save the page */
            sprintf (file_name, "%s/%s%1d.obj", bootDir, dumpName, page_num);
            SetCurDir (file_name);
            curFileDefined = TRUE;
            unlink (file_name);
            if (short_name = IsPrefix (bootDir, file_name, &rest)) ++rest;
            if ((page_fp = fopen (file_name, "w")) == NULL)
            {
               if (short_name)
                  printf ("Can not open '%s' for write.  Abort!\n", rest);
               else
                  printf ("Can not open '%s' for write.  Abort!\n", file_name);

               strcpy (curSymDir, saved_sym_dir);
               sprintf (file_name, "%s/%s", saved_dir_name, saved_file_name);
               SetCurDir (file_name);
               curFileDefined = saved_file_defined;
               fileModified = saved_file_modified;
               topObj = saved_top_obj;
               botObj = saved_bot_obj;
               ClearAndRedrawDrawWindow ();
               RedrawTitleWindow ();
               return;
            }
            if (short_name)
               printf ("Saving %1d objects in %s...\n", count, rest);
            else
               printf ("Saving %1d objects in %s...\n", count, file_name);
            Save (page_fp, botObj, 0);
            fclose (page_fp);

            DelAllObj ();
            ClearAndRedrawDrawWindow ();

            ClearFileInfo ();
            for (obj_ptr_2 = secondPageBotObj; obj_ptr_2 != NULL;
                  obj_ptr_2 = obj_ptr_2->prev)
            {
               if (obj_ptr_2!=secondPageTopAnchor &&
                     obj_ptr_2!=secondPageBotAnchor)
               {
                  obj_ptr_1 = DupObj (obj_ptr_2);
                  if (obj_ptr_1->type == OBJ_ICON)
                  {
                     struct AttrRec	* attr_ptr;

                     for (attr_ptr=obj_ptr_1->lattr; attr_ptr!=NULL;
                           attr_ptr=attr_ptr->prev)
                        if (strcmp (attr_ptr->name, "Page=") == 0)
                        {
                           sprintf (attr_ptr->s, "p%1d", page_num+1);
                           UpdAttr (attr_ptr);
                           break;
                        }
                  }
                  AdjObjSplineVs (obj_ptr_1);
                  AddObj (NULL, topObj, obj_ptr_1);
                  if (PointInBBox (obj_ptr_1->x, obj_ptr_1->y, drawWinBBox) ||
                        BBoxIntersect (obj_ptr_1->bbox, drawWinBBox))
                     DrawObj (drawWindow, obj_ptr_1);
               }
            }
            SetFileModified (TRUE);
            RedrawTitleWindow ();

            total += count;
            count = 0;
            page_num++;
            cur_y = secondPageTopAnchor->obbox.lty;
            bot_anchor = secondPageBotAnchor;
         }
      }
      else
         next_obj = obj_ptr->prev;
   }

   printf ("Total of %1d objects printed.\n", total);

   strcpy (curSymDir, saved_sym_dir);
   sprintf (file_name, "%s/%s", saved_dir_name, saved_file_name);
   SetCurDir (file_name);
   fileModified = saved_file_modified;
   curFileDefined = saved_file_defined;
   topObj = saved_top_obj;
   botObj = saved_bot_obj;
   ClearAndRedrawDrawWindow ();
   RedrawTitleWindow ();
}

static
void GenTsaiHsing ()
{
   TieLooseEnds ();
   SetCurChoice (NOTHING);
   if (topSel!=NULL) { HighLightReverse (); RemoveAllSel (); }

   GetXDefaults ();
   if (!ProcessTemplates ()) return;

   DumpTsaiHsing ();
}

static
int GetKeys (obj_ptr, key1, key2, key3)
   struct ObjRec	* obj_ptr;
   char			* key1, * key2, * key3;
{
   struct AttrRec	* attr_ptr;

   for (attr_ptr=obj_ptr->lattr; attr_ptr!=NULL; attr_ptr=attr_ptr->prev)
      if (strcmp (attr_ptr->name, firstKey) == 0)
      {
         strcpy (key1, attr_ptr->s);
         if (*key1 == '\0') return (FALSE);
         break;
      }
   if (*secondKey == '\0')
      *key2 = '\0';
   else
   {
      for (attr_ptr=obj_ptr->lattr; attr_ptr!=NULL; attr_ptr=attr_ptr->prev)
         if (strcmp (attr_ptr->name, secondKey) == 0)
         {
            strcpy (key2, attr_ptr->s);
            break;
         }
      if (attr_ptr == NULL) *key2 = '\0';
   }
   if (*thirdKey == '\0')
      *key3 = '\0';
   else
   {
      for (attr_ptr=obj_ptr->lattr; attr_ptr!=NULL; attr_ptr=attr_ptr->prev)
         if (strcmp (attr_ptr->name, thirdKey) == 0)
         {
            strcpy (key3, attr_ptr->s);
            break;
         }
      if (attr_ptr == NULL) *key3 = '\0';
   }
   return (TRUE);
}

static
int Hash (key)
   char	* key;
{
   if (*key >= 'A' && *key <= 'Z') return (*key-'A');
   return (26);
}

#define LESSTHAN -1
#define EQUAL 0
#define GREATERTHAN 1

static
int Compare (ptr1, ptr2)
   struct HashRec	* ptr1, * ptr2;
   /* returns EQUAL if ptr1==ptr2 */
   /* returns LESSTHAN if ptr1<ptr2 */
   /* returns GREATERTHAN if ptr1>ptr2 */
{
   register char	* c_ptr1, * c_ptr2;

   c_ptr1=ptr1->key1; c_ptr2=ptr2->key1;
   while (*c_ptr1 == *c_ptr2) { c_ptr1++; c_ptr2++; }
   if (*c_ptr1 != *c_ptr2)
      return ((*c_ptr1>*c_ptr2) ? GREATERTHAN : LESSTHAN);

   c_ptr1=ptr1->key2; c_ptr2=ptr2->key2;
   while (*c_ptr1 == *c_ptr2) { c_ptr1++; c_ptr2++; }
   if (*c_ptr1 != *c_ptr2) return ((*c_ptr1>*c_ptr2) ? GREATERTHAN : LESSTHAN);

   c_ptr1=ptr1->key3; c_ptr2=ptr2->key3;
   while (*c_ptr1 == *c_ptr2) { c_ptr1++; c_ptr2++; }
   if (*c_ptr1 != *c_ptr2) return ((*c_ptr1>*c_ptr2) ? GREATERTHAN : LESSTHAN);

   return (EQUAL);
}

static
void Sort ()
{
   char			key1[MAXSTRING], key2[MAXSTRING], key3[MAXSTRING];
   char			msg[MAXSTRING];
   int			i, cur_y, bucket, status;
   int			count, na_count, bad_key_count;
   struct ObjRec	* obj_ptr, * obj_ptr_1, * obj_with_keys;
   struct HashRec	* hash_ptr, * ptr;

   count = na_count = bad_key_count = 0;
   for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      count++;
      if ((obj_with_keys = DesiredObj (obj_ptr)) != NULL)
      {
         if (GetKeys (obj_with_keys, key1, key2, key3))
         {
            hash_ptr = (struct HashRec *) calloc (1, sizeof (struct HashRec));
            strcpy (hash_ptr->key1, key1);
            strcpy (hash_ptr->key2, key2);
            strcpy (hash_ptr->key3, key3);
            hash_ptr->value = obj_ptr;

            bucket = Hash (key1);
            for (ptr = hashBotArray[bucket]; ptr != NULL; ptr = ptr->prev)
            {
               status = Compare (ptr, hash_ptr);
               if (status == GREATERTHAN)
                  break;
            }
            hash_ptr->prev = ptr;
            if (ptr == NULL)
            {
               hash_ptr->next = hashTopArray[bucket];
               if (hashBotArray[bucket] == NULL)
                  hashBotArray[bucket] = hash_ptr;
               else
                  hashTopArray[bucket]->prev = hash_ptr;
               hashTopArray[bucket] = hash_ptr;
            }
            else
            {
               hash_ptr->next = ptr->next;
               if (ptr->next == NULL)
                  hashBotArray[bucket] = hash_ptr;
               else
                  ptr->next->prev = hash_ptr;
               ptr->next = hash_ptr;
            }
         }
         else
            bad_key_count++;
      }
      else
         na_count++;
   }
   if (bad_key_count != 0)
   {
      sprintf (msg, "%1d out of %1d objects have bad keys.", bad_key_count,
            count);
      printf ("%s\n", msg);
      Msg (msg);
   }
   if (na_count != 1)
   {
      sprintf (msg, "%1d out of %1d objects do not have attributes.", na_count,
            count);
      printf ("%s\n", msg);
      Msg (msg);
   }

   topObj = botObj = NULL;
   ClearAndRedrawDrawWindow ();

   for (obj_ptr = firstPageBotObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      if (obj_ptr != firstPageTopAnchor && obj_ptr != firstPageBotAnchor)
      {
         obj_ptr_1 = DupObj (obj_ptr);
         AdjObjSplineVs (obj_ptr_1);
         AddObj (NULL, topObj, obj_ptr_1);
         if (PointInBBox (obj_ptr_1->x, obj_ptr_1->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr_1->bbox, drawWinBBox))
            DrawObj (drawWindow, obj_ptr_1);
      }
   }

   cur_y = firstPageTopAnchor->obbox.lty;
   count = 0;

   for (i = 0; i < 27; i++)
   {
      for (hash_ptr=hashBotArray[i]; hash_ptr!=NULL; hash_ptr=hash_ptr->prev)
      {
         obj_ptr = hash_ptr->value;
         MoveObj (obj_ptr, firstPageTopAnchor->obbox.ltx-obj_ptr->obbox.ltx,
               cur_y-obj_ptr->obbox.lty);
         AddObj (NULL, topObj, obj_ptr);
         if (PointInBBox (obj_ptr->x, obj_ptr->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr->bbox, drawWinBBox))
            DrawObj (drawWindow, obj_ptr);

         cur_y += obj_ptr->obbox.rby-obj_ptr->obbox.lty;
         count++;
      }
   }
   sprintf (msg, "%1d objects sorted.", count);
   printf ("%s\n", msg);
   Msg (msg);
   SetFileModified (TRUE);
   RedrawTitleWindow ();
}

static
void SortTsaiHsing ()
{
   TieLooseEnds ();
   SetCurChoice (NOTHING);
   if (topSel!=NULL) { HighLightReverse (); RemoveAllSel (); }

   StartCompositeCmd ();
   SelectEverything ();
   PrepareToRecord (CMD_DELETE, topSel, botSel, numObjSelected);
   RecordCmd (CMD_DELETE, NULL, NULL, NULL, 0);

   GetXDefaults ();
   if (!ProcessTemplates ()) return;

   CleanUpHashTbl ();
   Sort ();

   SelectEverything ();
   PrepareToRecord (CMD_NEW, NULL, NULL, 0);
   RecordCmd (CMD_NEW, NULL, topSel, botSel, numObjSelected);
   RemoveAllSel ();
   EndCompositeCmd ();
}

static
void LastNameLast (RightOrder, Name)
   char	* RightOrder, * Name;
{
   register char	* c_ptr;
   char			name[MAXSTRING];

   strcpy (name, Name);
   for (c_ptr=name; *c_ptr!='\0' && *c_ptr!=','; c_ptr++) ;

   if (*c_ptr == '\0')
      strcpy (RightOrder, name);
   else
   {
      *c_ptr = '\0';
      for (c_ptr++; *c_ptr==' '; c_ptr++) ;

      if (*c_ptr == '\0')
         strcpy (RightOrder, name);
      else
         sprintf (RightOrder, "%s %s", c_ptr, name);
   }
}

static
int StateZipOk (s)
   register char	* s;
{
   register int		i;
   register char	* c_ptr;

   if (numStateConstraints == 0) return (TRUE);
   for (c_ptr=s; *c_ptr!='\0'; c_ptr++)
      if (*c_ptr >= '0' && *c_ptr <= '9')
      {
         int	and_condition=TRUE;

         for (i=0; i < numStateConstraints; i++)
         {
            if (strcmp (stateConstraints[i], "||") == 0)
            {
               if (and_condition)
                  return (TRUE);
               else
               {
                  and_condition = TRUE;
                  continue;
               }
            }
            else
               if (!and_condition) continue;
            if (*stateConstraints[i] == '~')
            {
               if (strncmp (s, &(stateConstraints[i])[1],
                     lenStateConstraints[i]-1) == 0)
                  and_condition = FALSE;
            }
            else
            {
               if (strncmp (s, stateConstraints[i],
                     lenStateConstraints[i]) != 0)
                  and_condition = FALSE;
            }
         }
         return (and_condition);
      }
   return (FALSE); /* ZIP code is required for label generation */
}

static
void GenLabel ()
{
   struct ObjRec	* obj_ptr, * obj_with_keys;
   struct AttrRec	* attr_ptr;
   int			has_name, has_addr, has_city, has_state_zip, count=0;
   char			name[MAXSTRING], addr[MAXSTRING], city[MAXSTRING];
   char			state_zip[MAXSTRING], msg[MAXSTRING];

   TieLooseEnds ();
   SetCurChoice (NOTHING);
   if (topSel!=NULL) { HighLightReverse (); RemoveAllSel (); }

   GetXDefaults ();

   if ((fp = fopen ("Labels", "w")) == NULL)
   {
      sprintf (msg, "Can not open 'Labels' for write.");
      printf ("%s\n", msg);
      Msg (msg);
      return;
   }

   for (obj_ptr=botObj; obj_ptr!=NULL; obj_ptr=obj_ptr->prev)
   {
      if ((obj_with_keys = DesiredObj (obj_ptr)) != NULL)
      {
         has_name = has_addr = has_city = has_state_zip = FALSE;

         attr_ptr = obj_with_keys->lattr;
         for ( ; attr_ptr!=NULL; attr_ptr=attr_ptr->prev)
         {
            if (strcmp (attr_ptr->name, "Name=") == 0)
            {
               LastNameLast (name, attr_ptr->s);
               has_name = TRUE;
            }
            else if (strcmp (attr_ptr->name, "Address=") == 0)
            {
               strcpy (addr, attr_ptr->s);
               has_addr = TRUE;
            }
            else if (strcmp (attr_ptr->name, "City=") == 0)
            {
               strcpy (city, attr_ptr->s);
               has_city = TRUE;
            }
            else if (strcmp (attr_ptr->name, "StateZip=") == 0 &&
                  StateZipOk(attr_ptr->s))
            {
               strcpy (state_zip, attr_ptr->s);
               has_state_zip = TRUE;
            }
            if (has_name && has_addr && has_city && has_state_zip) break;
         }
         if (has_name && has_addr && has_city && has_state_zip)
         {
            fprintf (fp, "%s\n%s\n%s, %s\n\n", name, addr, city, state_zip);
            count++;
         }
      }
   }
   fclose (fp);
   sprintf (msg, "%1d label entries generated.", count);
   printf ("%s\n", msg);
   Msg (msg);
}

#define SEARCH_BACKWARD 0
#define SEARCH_FORWARD 1

char	prevSearchStr[MAXSTRING+1];
int	prevSearchDirection=INVALID;

static
void Search (search_forward, again)
   int	search_forward, again;
{
   struct ObjRec	* obj_ptr, * obj_with_keys;
   struct AttrRec	* attr_ptr;
   char			search_str[MAXSTRING+1], msg[MAXSTRING+1];
   int			search_str_len;

   if (topSel != NULL && topSel != botSel)
   {
      Msg ("Can only search while 0 or 1 object is selected.");
      return;
   }

   if (again)
   {
      strcpy (search_str, prevSearchStr);
      if (*search_str == '\0') return;
   }
   else
   {
      Dialog ("Please enter an attribute to search:",
            "( <CR>: accept, <ESC>: cancel )", search_str);
      if (*search_str == '\0') return;
      strcpy (prevSearchStr, search_str);
   }
   search_str_len = strlen (search_str);
   prevSearchDirection = search_forward;

   if (curChoice != NOTHING)
   {
      TieLooseEnds ();
      SetCurChoice (NOTHING);
      if (topSel!=NULL) { HighLightReverse (); RemoveAllSel (); }
   }

   if (topSel == NULL)
      obj_ptr = (search_forward) ? botObj : topObj;
   else
      obj_ptr = (search_forward) ? topSel->obj->prev : topSel->obj->next;

   if (topSel!=NULL) { HighLightReverse (); RemoveAllSel (); }

   for ( ; obj_ptr!=NULL;
         obj_ptr=((search_forward) ? obj_ptr->prev : obj_ptr->next))
   {
      if ((obj_with_keys = DesiredObj (obj_ptr)) != NULL)
      {
         attr_ptr = obj_with_keys->lattr;
         for ( ; attr_ptr!=NULL; attr_ptr=attr_ptr->prev)
         {
            int	name_len = strlen (attr_ptr->name);

            if (search_str_len >= name_len &&
                  strncmp(attr_ptr->name, search_str, name_len) == 0)
            {
               if (strncmp(attr_ptr->s, &search_str[name_len],
                     search_str_len-name_len) == 0)
               {
                  if (topSel == NULL)
                  {
                     topSel = botSel = (struct SelRec *) calloc (1,
                           sizeof(struct SelRec));
                     topSel->next = topSel->prev = NULL;
                  }
                  topSel->obj = obj_ptr;
                  UpdSelBBox ();
                  HighLightForward ();
                  justDupped = FALSE;
                  if (!BBoxIntersect (obj_ptr->bbox, drawWinBBox))
                     PreciseZoom (zoomedIn, zoomScale, TRUE);
                  return;
               }
            }
         }
      }
   }
   sprintf (msg, "Can not find '%s'.", search_str);
   Dialog (msg, "( <CR> or <ESC> to continue )", search_str);
   if (topSel!=NULL) { HighLightReverse (); RemoveAllSel (); }
}

void SearchAgain ()
{
   char	dummy[MAXSTRING+1];

   if (prevSearchDirection == INVALID)
      Dialog ("No previous search!", "( <CR> or <ESC> to continue )", dummy);
   else
      Search (prevSearchDirection, TRUE);
}

#define TSAIHSING_READ 0
#define TSAIHSING_SORT 1
#define TSAIHSING_GEN 2
#define TSAIHSING_FILTER 3
#define TSAIHSING_LABEL 4
#define TSAIHSING_SEARCH_FORWARD 5
#define TSAIHSING_SEARCH_BACKWARD 6
#define TSAIHSING_SEARCH_AGAIN 7

#define MAXTSAIHSINGMENUS 8

static char	* tsaiHsingMenuStr[] =
{  "ReadExcel",
   "SortDrawing",
   "OutputMultiPage",
   "FilterDrawing",
   "GenerateLabel",
   "SearchForward",
   "SearchBackward",
   "SearchAgain"
};

void TsaiHsing ()
{
   int	index, * fore_colors, * valid, *init_rv, x, y;
   char	* c_ptr;

/* if ((c_ptr = getenv("USER")) == NULL) return; */
/* if (strcmp(c_ptr,"william") != 0) return; */

   CornerLoop (&x, &y);
   DefaultColorArrays (MAXTSAIHSINGMENUS, &fore_colors, &valid, &init_rv);
   index = TextMenuLoop (x, y, tsaiHsingMenuStr, MAXTSAIHSINGMENUS,
         fore_colors, valid, init_rv, SINGLECOLOR);
   switch (index)
   {
      case TSAIHSING_READ: ReadTsaiHsing (); break;
      case TSAIHSING_SORT: SortTsaiHsing (); break;
      case TSAIHSING_GEN: GenTsaiHsing (); break;
      case TSAIHSING_FILTER: FilterDrawing (); break;
      case TSAIHSING_LABEL: GenLabel (); break;
      case TSAIHSING_SEARCH_FORWARD: Search (SEARCH_FORWARD, FALSE); break;
      case TSAIHSING_SEARCH_BACKWARD: Search (SEARCH_BACKWARD, FALSE); break;
      case TSAIHSING_SEARCH_AGAIN: SearchAgain (); break;
   }
   return;
}
